"""Sample data for arcos_gui. Will be added to the napari sample data menu."""

from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

import requests
from arcos_gui import _main_widget
from arcos_gui.processing import columnnames
from napari import current_viewer
from napari.utils.notifications import show_info
from skimage.io import imread
from tqdm import tqdm

if TYPE_CHECKING:
    import napari.viewer


def open_plugin(viewer: napari.Viewer):
    """Main function. Adds plugin dock widget to napari viewer.

    Parameters
    ----------
    viewer : napari.Viewer
        Napari viewer to add the plugin to.

    Returns
    -------
    plugin : _main_widget.MainWindow
        The plugin instance.
    """
    plugin: _main_widget.MainWindow
    viewer, plugin = viewer.window.add_plugin_dock_widget(
        plugin_name="arcos-gui", widget_name="ARCOS Main Widget"
    )
    return plugin


def resolve(name, basepath=None):
    """Resolve path to file in the package directory."""
    if not basepath:
        basepath = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(basepath, name)


def download(url: str, fname: str):
    """Download file from url and save it to fname."""
    resp = requests.get(url, stream=True, timeout=10)
    curr_request = requests.head(url, timeout=10)
    if "text/html" in curr_request.headers["content-type"]:
        raise ValueError("URL is not a file")
    total = int(resp.headers.get("content-length", 0))
    with open(fname, "wb") as file, tqdm(
        desc=fname,
        total=total,
        unit="b",
        unit_scale=True,
        unit_divisor=1024,
    ) as pbar:
        for data in resp.iter_content(chunk_size=1024):
            size = file.write(data)
            pbar.update(size)


def load_synthetic_dataset(plugin: _main_widget.MainWindow | None = None):
    """Load sample data into stored_variables"""
    if plugin is None:
        plugin = _main_widget.MainWindow.get_last_instance()
    if not plugin:
        _plugin = open_plugin(current_viewer())
    else:
        _plugin = plugin

    sample_data_path = str(resolve(Path("arcos_data.csv")))
    columns = columnnames(
        frame_column="t",
        position_id="None",
        object_id="id",
        x_column="x",
        y_column="y",
        z_column="None",
        measurement_column_1="m",
        measurement_column_2="None",
        additional_filter_column="None",
        measurement_math_operation="None",
        measurement_bin=None,
        measurement_resc=None,
        collid_name="collid",
        measurement_column="m",
    )

    try:
        _plugin._input_controller.load_sample_data(sample_data_path, columns)
        _plugin._layer_prop_controller.widget.LUT.setCurrentText("viridis")
    except RuntimeError:
        show_info("Cannot find the plugin, please open it first")
    return []


def load_real_dataset(
    load_image: bool = True, plugin: _main_widget.MainWindow | None = None
):
    """Load sample data and get image correspoinding to the data"""
    if plugin is None:
        plugin = _main_widget.MainWindow.get_last_instance()
    if not plugin:
        _plugin = open_plugin(current_viewer())
    else:
        _plugin = plugin
    image_url = "https://macdobry.net/ARCOS/data/MDCK_example_event.tif"
    img_path = resolve("MDCK_example_event.tif")
    if not os.path.exists(img_path) and load_image:
        try:
            show_info("Image not found: Downloading example image")
            download(image_url, img_path)
        except Exception:
            show_info("Image not found: Downloading example image failed")

    if os.path.exists(img_path):
        img = imread(img_path)
        img_data_tuple = [(img, {"name": "Example ERK Wave", "colormap": "inferno"})]
    else:
        img_data_tuple = []

    columns = columnnames(
        frame_column="t",
        position_id="None",
        object_id="id",
        x_column="x",
        y_column="y",
        z_column="None",
        measurement_column_1="m",
        measurement_column_2="None",
        additional_filter_column="None",
        measurement_math_operation="None",
        measurement_bin=None,
        measurement_resc=None,
        collid_name="collid",
        measurement_column="m",
    )
    try:
        sample_data_path = str(resolve(Path("arcos_data_2.csv")))
        _plugin._input_controller.load_sample_data(sample_data_path, columns)
    except RuntimeError:
        show_info("Cannot find the plugin, please open it first")
    return img_data_tuple



"""Various utility functions for generating shapes and text."""
from __future__ import annotations

import numpy as np
import pandas as pd
from scipy.spatial import ConvexHull, QhullError

from ._config import ARCOS_LAYERS, COLOR_CYCLE


def reshape_by_input_string(selected_columns, input_string, vColsCore):
    """Reshape a 2D array of selected columns based on an input string.

    Parameters:
        selected_columns (np.ndarray): A 2D array of selected columns.
        Order Corresponds to vColsCore order.
        input_string (str): A string that contains the characters 't', 'x', 'y', and 'z'.
        vColsCore (list): List of column names in the order tyxz.

    Returns (np.ndarray): A 2D array with the selected columns reshaped based on the input string.
    """
    if input_string is None and len(vColsCore) == 3:
        # 2D data
        input_string = "tyx"

    elif input_string is None and len(vColsCore) == 4:
        # 3D data
        input_string = "tzyx"

    # Convert the input string to lowercase
    input_string = input_string.lower()

    # Check if the input string contains only the allowed characters
    if any(char not in "tzyx" for char in input_string):
        raise ValueError(
            "Input string must contain only the characters 't', 'x', 'y', and 'z'."
        )

    # Determine the length of the second dimension based on the input string
    second_dim_length = len(input_string)

    # Create a new array with zeros and the shape based on the input string
    reshaped_array = np.zeros((selected_columns.shape[0], second_dim_length))

    # Map the characters to the corresponding column indexes
    char_to_index = {"t": 0, "y": 1, "x": 2, "z": 3 if len(vColsCore) > 3 else None}

    # Fill the new array with the selected columns and zeros
    for i, char in enumerate(input_string):
        if (
            char_to_index[char] is not None
            and char_to_index[char] < selected_columns.shape[1]
        ):
            reshaped_array[:, i] = selected_columns[:, char_to_index[char]]

    return reshaped_array


def calculate_convex_hull(array):
    """Calculates the convex hull for a 2d array of points.

    Parameters:
        array (np.ndarray): A 2d array of points with x and y coordinates.

    Returns (np.ndarray): If shape of input array can be used
    to calculate convex hull the vertices of the convex hull are returned.
    If shape is less, the points themselfs are returned.
    """
    # Maybe there is a better check for coplanar
    # points rathern than using QhullError
    try:
        if array.shape[0] > 2:
            hull = ConvexHull(array[:, 2:])
            array_out = array[hull.vertices]
            return array_out
        if array.shape[0] == 2:
            return array
        return np.array([])
    except QhullError:
        return np.array([])


def calculate_convex_hull_3d(array):
    """Calculates the convex hull for a 3d array of points.

    Parameters:
        array (np.ndarray): A 2d array of points with x y and z coordinates.

    Returns (np.ndarray): If shape of input array can be used
    to calculate convex hull the vertices of the convex hull are returned.
    If shape is less, the points themselfs are returned.
    """
    if array.shape[0] > 3:
        hull = ConvexHull(array[:, 2:])
        array_faces = hull.simplices
        if array_faces.shape[0] != 0:
            return array_faces
    if array.shape[0] == 3:
        array_faces = np.array([[0, 1, 2]])
        return array_faces
    if array.shape[0] == 2:
        array_faces = np.array([[0, 1, 1]])
        return array_faces
    if array.shape[0] == 1:
        array_faces = np.array([[0, 0, 0]])
        return array_faces


def get_verticesHull(
    df,
    frame,
    colid,
    col_x,
    col_y,
):
    """Calculate convex hull for 2d collective events.

    Input dataframe is converted into a numpy array and split into groups
    according to unique collective ids.
    For each set array the convex hull is calculated.

    Parameters:
        df (pd.DataFrame): A dataframe with at least 4 columns
        containing the XY coordinates of a set of points aswell
        as frame and collective id columns.
        frame (str): Name of frame column in df.
        colid (str): Name of collective id column in df.
        col_x (str): Name of column x coordinate column in df.
        col_y (str): Name of column y coordinate column in df.

    Returns (np.ndarray,np.ndarray): Tuple containing arrays of vertices,
    one for each collective event. Array of colors, a unique
    one for each collective event.
    """

    df = df.sort_values([colid, frame])
    array_txy = df[[colid, frame, col_y, col_x]].to_numpy()
    array_txy = array_txy[~np.isnan(array_txy).any(axis=1)]
    grouped_array = np.split(
        array_txy, np.unique(array_txy[:, 0:2], axis=0, return_index=True)[1][1:]
    )
    # map to grouped_array
    convex_hulls = [calculate_convex_hull(i) for i in grouped_array if i.shape[0] > 1]
    color_ids = np.take(
        np.array(COLOR_CYCLE),
        [int(i[0, 0]) for i in convex_hulls if i.size > 0],
        mode="wrap",
    )
    # color_ids = recycle_palette(COLOR_CYCLE, len(convex_hulls))
    out = [i[:, 1:] for i in convex_hulls if i.size > 0]
    return out, color_ids


# @profile
def make_surface_3d(
    df: pd.DataFrame,
    vColsCore: list[str],
    colid: str,
    output_order: str | None = "tyxz",
):
    """Calculate convex hull for 3d collective events.

    Input dataframe is converted into a numpy array and split into groups
    according to unique collective ids.
    For each set array the convex hull is calculated.
    A tuple that can be used to add 3d convex hull with the napari
    add_surface function is generated.
    Output has to be appended with empy vertices and surfaces for the
    timepoints where no surface should be drawn. Otherwise will
    result in a nontype subscription error.

    Parameters:
        df (pd.DataFrame): A dataframe with at least 4 columns
        containing the XY coordinates of a set of points aswell
        as frame and collective id columns.
        frame (str): Name of frame column in df.
        colid (str): Name of collective id column in df.
        vColsCore (list): List of column names in the order tyxz.
        output_order (str): String that defines the order of the output. Default is tyxz.

    Returns (tuple(np.ndarray, np.ndarray, np.ndarray)): Tuple that contains
    vertex coordinates, face indices and color ids
    """
    if output_order is None:
        output_order = "tyxz"
    frame = vColsCore[0]

    # reorder columns according to output_order
    map_dict = {
        "t": vColsCore[0],
        "y": vColsCore[1],
        "x": vColsCore[2],
        "z": vColsCore[3],
    }
    reordered_cols = [map_dict[i] for i in output_order]

    data_faces = []
    vertices_count = 0
    # sort needed for np.split
    df = df.sort_values([colid, frame])
    selection_cols = [colid] + reordered_cols
    array_idtyxz = df[selection_cols].to_numpy()
    array_idtyxz = array_idtyxz[~np.isnan(array_idtyxz).any(axis=1)]
    # split array into list of arrays, one for each collid/timepoint combination
    grouped_array = np.split(
        array_idtyxz, np.unique(array_idtyxz[:, 0:2], axis=0, return_index=True)[1][1:]
    )
    # calc convex hull for every array in the list
    convex_hulls = [calculate_convex_hull_3d(i) for i in grouped_array]
    # generates color ids (integers for LUT in napari)
    color_ids = np.concatenate([i[:, 0].astype(np.int64) for i in grouped_array])
    out_vertices = np.concatenate(grouped_array)[:, 1:]
    # merge convex hull face list and shift indexes according to groups
    for i, val in enumerate(convex_hulls):
        data_faces.append(np.add(val, vertices_count))
        vertices_count += len(grouped_array[i])
    out_faces = np.concatenate(data_faces)
    return (out_vertices, out_faces, color_ids)


# @profile
def fix_3d_convex_hull(df, vertices, faces, colors, col_t):
    """Generate empty vertex and faces to fix napari subset error.

    Parameters:
        df (pd.DataFrame): A dataframe used to calculate convex hulls.
        vertices (np.ndarray): vertex coordinates.
        faces (np.ndarray): Array containing face indices.
        colors (np.ndarray): Array containing color ids.
        col_t (str): String name of frame column in df.
    """
    empty_vertex = []
    empty_faces = []
    empty_colors = []
    time_points = np.unique(vertices[:, 0])
    arr_size = vertices.shape[0]
    for i in df[col_t].unique():
        if i not in time_points:
            empty_vertex.append([i, 0, 0, 0])
            empty_faces.append([arr_size, arr_size, arr_size])
            arr_size = arr_size + 1
            empty_colors.append(0)

    empty_vertex = np.array(empty_vertex)
    empty_faces = np.array(empty_faces)
    empty_colors = np.array(empty_colors)

    if empty_vertex.size > 0:
        vertices = np.concatenate((vertices, empty_vertex), axis=0)
        faces = np.concatenate((faces, empty_faces), axis=0)
        colors = np.concatenate((colors, empty_colors), axis=0)
        return (vertices, faces, colors)
    return (vertices, faces, colors)


def calc_bbox(array: np.ndarray):
    """Calcualte the bounding box of input array.

    Parameters:
        array (np.ndarray): 2d array of coordinates
        for which to compute the bounding box.

    Returns (np.ndarray): 2d array of coordinates for the bounding box.
    """
    # reorder columns according to output_order

    time_point = array[0, 0]
    pos_array = array[:, 1:]
    # 3d case
    if pos_array.shape[1] == 3:
        miny, minx, minz = np.min(pos_array, axis=0)
        maxy, maxx, maxz = np.max(pos_array, axis=0)
        return np.array(
            [
                [time_point, miny, minx, minz],
                [time_point, miny, minx, maxz],
                [time_point, miny, maxx, maxz],
                [time_point, miny, maxx, minz],
                [time_point, maxy, maxx, minz],
                [time_point, maxy, minx, minz],
                [time_point, maxy, minx, maxz],
                [time_point, maxy, maxx, maxz],
            ]
        )
    # 2d case
    miny, minx = np.min(pos_array, axis=0)
    maxy, maxx = np.max(pos_array, axis=0)
    return np.array(
        [
            [time_point, miny, minx],
            [time_point, miny, maxx],
            [time_point, maxy, maxx],
            [time_point, maxy, minx],
        ]
    )


def get_bbox(
    df: pd.DataFrame,
    clid: int,
    frame: str,
    xcol: str,
    ycol: str,
    edge_size: float = 10,
    output_order: str = "tyx",
):
    """Get bounding box of dataframe in long format with position columns, for 2d case.

    Parameters:
        df (pd.DataFrame): dataframe to get bounding box form.
        frame (str): Name of frame column.
        xcol (str): X coordinate column.
        ycol (str): Y coordinate column.
        edge_size (float): Bounding Box edge_size, also used to calculate text size.

    Returns (nd.array, dict): Array that can be added to
    napari with add_shapes function aswell
    as dictionary that can be unpacked
    containing kwargs for shapes layer.
    """
    df = df.sort_values([frame])
    array_tpos = df[[frame, ycol, xcol]].to_numpy()
    array_tpos = array_tpos[~np.isnan(array_tpos).any(axis=1)]
    # reorder columns according to output_order
    # split array into list of arrays, one for each collid/timepoint combination
    grouped_array = np.split(
        array_tpos, np.unique(array_tpos[:, 0], axis=0, return_index=True)[1][1:]
    )
    # calc bbox for every array in the list
    bbox = [calc_bbox(i) for i in grouped_array]
    bbox = [reshape_by_input_string(i, output_order, [frame, ycol, xcol]) for i in bbox]
    text_size = edge_size * 2.5
    if text_size < 1:
        text_size = 1
    text_parameters = {
        "string": "Event Nbr: {label}",
        "size": text_size,
        "color": "white",
        "anchor": "upper_left",
        "translation": [-3, 0],
    }
    bbox_layer: dict = {}
    bbox_layer["properties"] = {}
    bbox_layer["properties"]["label"] = clid
    bbox_layer["text"] = text_parameters
    bbox_layer["face_color"] = "transparent"
    bbox_layer["edge_color"] = "red"
    bbox_layer["edge_width"] = edge_size
    bbox_layer["name"] = ARCOS_LAYERS["event_boundingbox"]

    return bbox, bbox_layer


def get_bbox_3d(
    df: pd.DataFrame,
    frame: str,
    xcol: str,
    ycol: str,
    zcol: str,
    output_order: str = "tzyx",
):
    """Get bounding box of dataframe in long format with position columns, for 3d case.

    Can be added to napari with the add_surfaces function.

    Parameters:
        df (pd.DataFrame): dataframe to get bounding box form.
        frame (str): Name of frame column.
        xcol (str): X coordinate column.
        ycol (str): Y coordinate column.
        zcol (str): Z coordinate column.

    Returns (nd.array, np.ndarray, np.ndarray): Tuple that
    can be added to napari with add_shapes
    function. Need to be passed on to the fix_3d_convex_hull
    function to avoid indexing errors in napari.
    """
    df = df.sort_values([frame])
    array_tpos = df[[frame, ycol, xcol, zcol]].to_numpy()
    array_tpos = array_tpos[~np.isnan(array_tpos).any(axis=1)]

    # split array into list of arrays, one for each collid/timepoint combination
    grouped_array = np.split(
        array_tpos, np.unique(array_tpos[:, 0], axis=0, return_index=True)[1][1:]
    )
    # calc bbox for every array in the list
    bbox = [calc_bbox(i) for i in grouped_array]
    bbox = [
        reshape_by_input_string(i, output_order, [frame, ycol, xcol, zcol])
        for i in bbox
    ]
    data_faces = []
    vertices_count = 0
    data_colors: np.ndarray = np.array([])
    # precalculated face indidec for a 3d bounding box
    face = np.array(
        [
            [3, 5, 4],
            [3, 5, 0],
            [3, 1, 2],
            [3, 1, 0],
            [7, 3, 2],
            [7, 3, 4],
            [6, 1, 0],
            [6, 5, 0],
            [6, 1, 2],
            [6, 7, 2],
            [6, 5, 4],
            [6, 7, 4],
        ]
    )
    for value in bbox:
        data_faces.append(np.add(face, vertices_count))
        vertices_count += len(value)
    out_faces = np.concatenate(data_faces)
    bbox_array = np.concatenate(bbox)
    data_colors = np.array([1 for i in range(bbox_array.shape[0])])
    return (bbox_array, out_faces, data_colors)



"""Module containing plotting classes."""

from re import findall
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import napari
import numpy as np
import pandas as pd
from arcos4py.tools import calculate_statistics
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
from napari.utils.notifications import show_info
from qtpy import QtWidgets
from scipy.stats import gaussian_kde

from ._config import ARCOS_LAYERS, COLOR_CYCLE
from ._shape_functions import fix_3d_convex_hull, get_bbox, get_bbox_3d

if TYPE_CHECKING:
    import napari.layers
    import napari.viewer


class CollevPlotter(QtWidgets.QWidget):
    """
    QWidget for plotting a scatterplot of Collective events.
    Make a matplotlib figure canvas and add it to a Qwidget.
    Canvas, figure and axis objects can be acessed by self.canvas,
    self.fig and self.ax. This plots duration of Collective events over their size as
    returned by arcos.

    Attributes:
        viewer (napari.viewer.Viewer): Napari viewer instance
        parent (qtpy.QtWidgets.QWidget): Parent widget, optional
    """

    def __init__(self, viewer: napari.viewer.Viewer, parent=None):
        """Class constructor.

        Parameters:
            viewer (napari.viewer.Viewer): Napari viewer instance
            parent (qtpy.QtWidgets.QWidget): Parent widget, optional
        """
        super().__init__(parent)
        self.viewer = viewer
        self.collid_name: str = "collid"
        self.nbr_collev: int = 0
        self.stats = pd.DataFrame(
            data={"total_size": [], "duration": [], self.collid_name: []}
        )
        self._callbacks: list = []
        self._init_mpl_widgets()
        self.fig.canvas.mpl_connect("motion_notify_event", self.hover)
        self.fig.canvas.mpl_connect("pick_event", self.on_pick)
        self.arcos = pd.DataFrame(data={"t": [], "id": [], "x": [], "y": [], "z": []})
        self.point_size = 10
        self.frame_col = "time"
        self.trackid_col = "id"
        self.posx = "x"
        self.posy = "y"
        self.posz = "z"
        self.output_order = "tzyx"
        self.update_plot(
            self.frame_col,
            self.trackid_col,
            self.posx,
            self.posy,
            self.posz,
            self.arcos,
            self.output_order,
        )

    def _init_mpl_widgets(self):
        """
        Method to initialise a matplotlib figure canvas, to generate,
        set plot style and axis, and populate it with a matplotlib.figure.Figure.
        """
        # construct layout
        self.layout_collevplot = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout_collevplot)

        # set up figure and axe objects
        with plt.style.context("dark_background"):
            plt.rcParams["axes.edgecolor"] = "#ffffff"
            self.fig = Figure(figsize=(2.5, 1.5))
            self.canvas = FigureCanvas(self.fig)
            self.ax = self.fig.add_subplot(111)
            self.ax.set_xlabel("Total Size")
            self.ax.set_ylabel("Event Duration")

        self.toolbar = NavigationToolbar(self.canvas, self)
        self.layout_collevplot.addWidget(self.toolbar)
        self.layout_collevplot.addWidget(self.canvas)
        self.setWindowTitle("Collective Events")
        self._update_layout()

    def _update_layout(self):
        self.fig.tight_layout(pad=0.1, w_pad=0.001, h_pad=0.05)

    def clear_plot(self):
        """Method to clear the plot."""
        self.ax.cla()
        self.ax.set_xlabel("Total Size")
        self.ax.set_ylabel("Event Duration")
        self.fig.canvas.draw_idle()

    def update_plot(
        self,
        frame_col,
        trackid_col,
        posx,
        posy,
        posz,
        arcos_data,
        arcos_stats,
        point_size=10,
        output_order="tzyx",
    ):
        """
        Method to update the matplotlibl axis object self.ax with new values from
        the stored_variables object
        """
        collev_stats = arcos_stats
        self.arcos = arcos_data
        self.point_size = point_size
        self.frame_col = frame_col
        self.trackid_col = trackid_col
        self.posx = posx
        self.posy = posy
        self.posz = posz
        self.output_order = output_order
        # if no calculation was run so far (i.e. when the widget is initialized)
        # populate it with no data else calculate stats for collective events,
        # generate plot with data
        if not self.arcos.empty and self.trackid_col:
            self.stats = collev_stats

            self.ax.cla()
            self.ax.spines["bottom"].set_color("white")
            self.ax.spines["top"].set_color("white")
            self.ax.spines["right"].set_color("white")
            self.ax.spines["left"].set_color("white")
            self.ax.xaxis.label.set_color("white")
            self.ax.yaxis.label.set_color("white")
            self.ax.tick_params(colors="white", which="both")
            self.ax.axis("on")
            self.ax.scatter(
                self.stats[["total_size"]],
                self.stats[["duration"]],
                alpha=0.8,
                picker=True,
            )
            self.ax.set_xlabel("Total Size")
            self.ax.set_ylabel("Event Duration")
            self.fig.canvas.draw_idle()
            self.nbr_collev = self.stats[self.collid_name].nunique()
        # generate empty annotation, set it to invisible
        self.annot = self.ax.annotate(
            "",
            xy=(0, 0),
            xytext=(0, 0),
            bbox=dict(boxstyle="round", fc="#252932", ec="white", linewidth=0.3),
            fontsize=7,
            color="white",
            clip_on=True,
        )
        self.annot.set_visible(False)
        # instantiate blitmanager and add annotation to it.
        # Used to improve performance of annotations
        # rendering annotation label.

        self.blit_manager = BlitManager(self.canvas, [self.annot])

    def update_annot(self, ind):
        """Update the annotation.

        Updates hover annotation showing collective event id.

        Parameters:
            ind (dict): Index of plotted data corresponding to the
            current mouse location.
        """
        pos = self.ax.collections[0].get_offsets()[ind["ind"][0]]
        pos_text = pos.copy()
        text = f"id: {int(self.stats[self.collid_name][ind['ind'][0]])}"
        self.annot.set_text(text)
        self.fig.canvas.draw_idle()
        renderer = self.fig.canvas.get_renderer()
        bbox = self.annot.get_window_extent(renderer)
        bbox_data = self.ax.transData.inverted().transform(bbox)
        xlim = self.ax.get_xlim()
        ylim = self.ax.get_ylim()
        size_h = bbox_data[1][0] - bbox_data[0][0]
        size_v = bbox_data[1][1] - bbox_data[0][1]

        # moves lable by the size of the bounding box
        # if it would be cut off by the plot boundry
        if pos_text[0] > (xlim[1] - size_h):
            pos_text[0] -= size_h
        if pos_text[1] > (ylim[1] - size_v):
            pos_text[1] -= size_v
        self.annot.xy = pos
        self.annot.set_position(pos_text)
        self.annot.get_bbox_patch().set_alpha(1)

    def hover(self, event):
        """Display annotation of collective even on hover.

        Checks if current mouse position is over a datapoint. If yes,
        displays corresponding collective event id.
        """
        vis = self.annot.get_visible()
        if event.inaxes == self.ax:
            try:
                cont, ind = self.ax.collections[0].contains(event)
                if cont:
                    self.update_annot(ind)
                    self.annot.set_visible(True)
                    # blitting to improve performance.
                    self.blit_manager.update()
                else:
                    if vis:
                        self.annot.set_visible(False)
                        # blitting to improve performance.
                        self.blit_manager.update()
            except IndexError:
                pass

    def on_pick(self, event):
        """Displays the selected collective event in the napari viewer.

        On pick gets collective event id from picked datapoint, gets
        the correspondig starting frame, moves to this and
        draws a bounding box arround the extends of the collective event.

        Parameters:
            event (matplotlib_pick_event): event generated from selecting a datapoint.
        """
        ind = event.ind
        clid = int(self.stats.iloc[ind[0]][self.collid_name])
        current_colev = self.arcos[self.arcos["collid"] == clid]
        edge_size = self.point_size / 5
        frame = self.stats.iloc[ind[0]].iloc[2]
        if ARCOS_LAYERS["event_boundingbox"] in self.viewer.layers:
            self.viewer.layers.remove(ARCOS_LAYERS["event_boundingbox"])
        if self.posz is None:
            bbox, bbox_param = get_bbox(
                current_colev,
                clid,
                self.frame_col,
                self.posx,
                self.posy,
                edge_size,
                self.output_order,
            )
            self.viewer.add_shapes(bbox, **bbox_param)
        else:
            timepoints = list(range(0, int(self.viewer.dims.range[0][1])))
            df_tp = pd.DataFrame(timepoints, columns=[self.frame_col])
            bbox_tuple = get_bbox_3d(
                current_colev,
                self.frame_col,
                self.posx,
                self.posy,
                self.posz,
                self.output_order,
            )
            bbox_tuple = fix_3d_convex_hull(
                df_tp, bbox_tuple[0], bbox_tuple[1], bbox_tuple[2], self.frame_col
            )
            self.viewer.add_surface(
                bbox_tuple,
                colormap="red",
                opacity=0.15,
                name=ARCOS_LAYERS["event_boundingbox"],
                shading="none",
            )

        current_step = list(self.viewer.dims.current_step)
        t_index = self.output_order.index("t")
        current_step[t_index] = frame
        self.viewer.dims.current_step = current_step


class NoodlePlot(QtWidgets.QWidget):
    """
    QWidget for plotting.
    Class to make a matplotlib figure canvas and add it to a Qwidget.
    Canvas, figure and axis objects can be acessed by self.canvas,
    self.fig and self.ax. This plots tracks of Collective events over time.
    """

    def __init__(self, viewer: napari.viewer.Viewer, parent=None):
        """
        QWidget for plotting a Noodleplot of Collective events.
        Tracks of objects are plotted and colored by collective event id.
        Make a matplotlib figure canvas and add it to a Qwidget.
        Canvas, figure and axis objects can be acessed by self.canvas,
        self.fig and self.ax.

        Attributes:
            viewer (napari.viewer.Viewer): Napari viewer instance
            parent (qtpy.QtWidgets.QWidget): Parent widget, optional
        """
        super().__init__(parent)
        self.viewer = viewer
        self.collid_name: str = "collid"
        self.nbr_collev: int = 0
        self.stats = pd.DataFrame(
            data={"total_size": [], "duration": [], self.collid_name: []}
        )
        self._callbacks: list = []
        self._init_mpl_widgets()
        self.arcos = pd.DataFrame(data={"t": [], "id": [], "x": [], "y": [], "z": []})
        self.point_size = 10
        self.frame_col = "frame"
        self.trackid_col = "id"
        self.posx = "x"
        self.posy = "y"
        self.posz = "z"
        self.projection_index = 3
        self.output_order = "tzyx"
        self.update_plot(
            self.frame_col,
            self.trackid_col,
            self.posx,
            self.posy,
            self.posz,
            self.arcos,
            self.output_order,
        )
        self.fig.canvas.mpl_connect("motion_notify_event", self.hover)
        self.fig.canvas.mpl_connect("pick_event", self.on_pick)
        self.combo_box.currentIndexChanged.connect(self.update_plot_data)

    def _init_mpl_widgets(self):
        """
        Method to initialise a matplotlib figure canvas, to generate,
        set plot style and axis, and populate it with a matplotlib.figure.Figure.
        """
        # construct layout
        self.combo_box = QtWidgets.QComboBox(self)
        combobox_label = QtWidgets.QLabel(self)
        combobox_label.setText("Projection Axis")
        layout_noodle_plot = QtWidgets.QVBoxLayout()
        layout_combobox = QtWidgets.QHBoxLayout()
        layout_combobox.addWidget(combobox_label)
        layout_combobox.addWidget(self.combo_box)
        layout_noodle_plot.addLayout(layout_combobox)

        # set up figure and axe objects
        with plt.style.context("dark_background"):
            plt.rcParams["axes.edgecolor"] = "#ffffff"
            self.fig = Figure(figsize=(2.5, 1.5))
            self.canvas = FigureCanvas(self.fig)
            self.ax = self.fig.add_subplot(111)
            self.ax.set_xlabel("Time")
            self.ax.set_ylabel("Position (px)")

        self.toolbar = NavigationToolbar(self.canvas, self)
        layout_noodle_plot.addWidget(self.toolbar)
        layout_noodle_plot.addWidget(self.canvas)
        self.setLayout(layout_noodle_plot)
        self.setWindowTitle("Noodle Plot")
        self._update_layout()

    def _update_layout(self):
        self.fig.tight_layout(pad=0.1, w_pad=0.001, h_pad=0.05)

    def prepare_data(
        self,
        df: pd.DataFrame,
        colev: str,
        trackid: str,
        frame: str,
        posx: str,
        posy: str,
        posz: str,
    ):
        """From arcos collective event data,
        generates a list of numpy arrays, one for each event.

        Parameters:
            df (pd.DataFrame): DataFrame containing collective events from arcos.
            colev (str): Name of the collective event column in df.
            trackid (str): Name of the track column in df.
            frame: (str): Name of the frame column in df.
            posx (str): Name of the X coordinate column in df.
            posy (str): Name of the Y coordinate column in df.
            posz (str): Name of the Z coordinate column in df,
            or "None" (str) if no z column.

        Returns (list[np.ndarray], np.ndarray): List of collective events data,
        colors for each collective event.
        """
        col_fact = f"{trackid}_factorized"

        # factorize column in order to prevent numpy grouping error in detrending
        value, _ = df[trackid].factorize()
        df[col_fact] = value

        # values need to be sorted to group with numpy
        df.sort_values([colev, col_fact], inplace=True)
        if posz:
            array = df[[colev, col_fact, frame, posx, posy, posz]].to_numpy()
        else:
            array = df[[colev, col_fact, frame, posx, posy]].to_numpy()
        # generate goroups for each unique value
        grouped_array = np.split(
            array, np.unique(array[:, 0], axis=0, return_index=True)[1][1:]
        )
        # make collids sequential
        seq_colids = np.concatenate(
            [np.repeat(i, value.shape[0]) for i, value in enumerate(grouped_array)],
            axis=0,
        )
        array_seq_colids = np.column_stack((array, seq_colids))
        # split sequential collids array by trackid and collid
        grouped_array = np.split(
            array_seq_colids,
            np.unique(array_seq_colids[:, :2], axis=0, return_index=True)[1][1:],
        )
        # generate colors for each collective event, wrap arround the color cycle
        colors = np.take(
            np.array(COLOR_CYCLE), [i + 1 for i in np.unique(seq_colids)], mode="wrap"
        )
        return grouped_array, colors

    def calc_stats(self, frame_col, trackid_col):
        """Calculates stats for collective events."""
        # if no calculation was run so far (i.e. when the widget is initialized)
        # populate it with no data
        if not self.arcos.empty:
            self.stats = calculate_statistics(
                self.arcos, frame_col, self.collid_name, trackid_col
            )

    def clear_plot(self):
        """Method to clear the plot."""
        self.ax.cla()
        self.ax.set_xlabel("Total Size")
        self.ax.set_ylabel("Event Duration")
        self.fig.canvas.draw_idle()

    def update_plot(
        self,
        frame_col,
        trackid_col,
        posx,
        posy,
        posz,
        arcos_data,
        point_size=10,
        output_order="tzyx",
    ):
        """
        Method to update the matplotlibl axis object self.ax with new values from
        the stored_variables object and update the projection choices.

        Parameters:
            frame_col (str): The name of the frame column.
            trackid_col (str): The name of the trackid column.
            posx (str): The name of the x coordinate column.
            posy (str): The name of the y coordinate column.
            posz (str): The name of the z coordinate column.
            arcos_data (pd.DataFrame): DataFrame containing arcos output data.
            point_size (int): The size of the points drawn in napari.
        """
        self.arcos = arcos_data
        self.point_size = point_size
        self.frame_col = frame_col
        self.trackid_col = trackid_col
        self.posx = posx
        self.posy = posy
        self.posz = posz
        self.output_order = output_order
        if self.posz:
            projection_list = [self.posx, self.posy, self.posz]
        else:
            projection_list = [self.posx, self.posy]
        self.combo_box.clear()
        self.combo_box.addItems(projection_list)
        self.update_plot_data()

    def update_plot_data(self):
        """Update plot data."""
        # if no calculation was run so far (i.e. when the widget is initialized)
        # populate it with no data
        projection_type = self.combo_box.currentText()
        if projection_type == self.posx:
            self.projection_index = 3
        elif projection_type == self.posy:
            self.projection_index = 4
        elif projection_type == self.posz:
            self.projection_index = 5
        if not self.arcos.empty and self.trackid_col:
            self.dat_grpd, self.colors = self.prepare_data(
                self.arcos,
                "collid",
                self.trackid_col,
                self.frame_col,
                self.posx,
                self.posy,
                self.posz,
            )
            self.calc_stats(self.frame_col, self.trackid_col)
            self.ax.cla()
            self.ax.spines["bottom"].set_color("white")
            self.ax.spines["top"].set_color("white")
            self.ax.spines["right"].set_color("white")
            self.ax.spines["left"].set_color("white")
            self.ax.xaxis.label.set_color("white")
            self.ax.yaxis.label.set_color("white")
            self.ax.tick_params(colors="white", which="both")
            self.ax.axis("on")
            self.ax.set_xlabel("Time Point")
            self.ax.set_ylabel("Position")
            self.fig.canvas.draw_idle()
            for dat in self.dat_grpd:
                self.ax.plot(
                    dat[:, 2],
                    dat[:, self.projection_index],
                    c=self.colors[int(dat[0, -1])],
                    picker=1,
                )

            self.nbr_collev = self.stats[self.collid_name].nunique()
        # generate empty annotation
        self.annot = self.ax.annotate(
            "",
            xy=(0, 0),
            xytext=(0, 0),
            bbox=dict(boxstyle="round", fc="#252932", ec="white", linewidth=0.3),
            fontsize=7,
            color="white",
            clip_on=True,
            animated=True,
        )
        self.annot.set_visible(False)
        # instantiate the BlitManager for faster rendering of the hover annotation.
        self.blint_manager = BlitManager(self.canvas, [self.annot])

    def update_annot(self, ind, line):
        """Update the annotation.

        Updates hover annotation showing collective event id.

        Parameters:
            ind (dict): Index of plotted data corresponding
            to the current mouse location.
            line (matplotlib.Artist.artist): Line artist where hover detected the event.
        """
        x_vals, y_vals = line.get_data()
        pos_text = [x_vals[ind["ind"][0]], y_vals[ind["ind"][0]]]
        clid_index = int(findall(r"\d+", line.get_label())[0])
        clid = int(self.dat_grpd[clid_index][0, 0])
        self.annot.xy = (x_vals[ind["ind"][0]], y_vals[ind["ind"][0]])
        text = f"id:{clid}"
        self.annot.set_text(text)
        # get size of the annotation bbox
        canvas_renderer = self.fig.canvas.get_renderer()
        bbox = self.annot.get_window_extent(renderer=canvas_renderer)
        bbox_data = self.ax.transData.inverted().transform(bbox)
        xlim = self.ax.get_xlim()
        ylim = self.ax.get_ylim()
        size_h = bbox_data[1][0] - bbox_data[0][0]
        size_v = bbox_data[1][1] - bbox_data[0][1]

        # fix for annotation drawn at the border of the plot (moves annotation)
        if pos_text[0] > (xlim[1] - size_h):
            pos_text[0] -= size_h
        if pos_text[1] > (ylim[1] - size_v):
            pos_text[1] -= size_v

        self.annot.xy = (x_vals[ind["ind"][0]], y_vals[ind["ind"][0]])
        self.annot.set_position(pos_text)
        self.annot.get_bbox_patch().set_alpha(1)

    def hover(self, event):
        """Display annotation of collective even on hover.

        Checks if current mouse position is over a datapoint. If yes,
        displays corresponding collective event id for the correct Line2D artist.
        """
        vis = self.annot.get_visible()
        if event.inaxes == self.ax:
            # get line, or liens that was hovered over
            selected_line = [line for line in self.ax.lines if line.contains(event)[0]]
            if selected_line:
                # loop over all lines that are within hover radius
                for line in selected_line:
                    _, ind = line.contains(event)
                    self.update_annot(ind, line)
                    self.annot.set_visible(True)
                    # blitting for faster rendering of annotations.
                    self.blint_manager.update()
                    break  # break on first line with content
            else:
                if vis:
                    self.annot.set_visible(False)
                    # blitting for faster rendering of annotations.
                    self.blint_manager.update()

    def on_pick(self, event):
        """Displays the selected collective event in the napari viewer.

        On pick gets collective event id from picked datapoint, gets
        the correspondig starting frame, moves to this and
        draws a bounding box arround the extends of the collective event.

        Parameters:
            event (matplotlib_pick_event): event generated from selecting a datapoint.
        """
        clid_index = int(findall(r"\d+", event.artist.get_label())[0])
        clid = int(self.dat_grpd[clid_index][0, 0])
        current_colev = self.arcos[self.arcos["collid"] == clid]
        edge_size = self.point_size / 5
        frame = int(
            self.stats[self.stats[self.collid_name] == clid]["first_timepoint"].iloc[0]
        )
        if ARCOS_LAYERS["event_boundingbox"] in self.viewer.layers:
            self.viewer.layers.remove(ARCOS_LAYERS["event_boundingbox"])
        if self.posz is None:
            bbox, bbox_param = get_bbox(
                current_colev,
                clid,
                self.frame_col,
                self.posx,
                self.posy,
                edge_size,
                self.output_order,
            )
            self.viewer.add_shapes(bbox, **bbox_param)
        else:
            timepoints = list(range(0, int(self.viewer.dims.range[0][1])))
            df_tp = pd.DataFrame(timepoints, columns=[self.frame_col])
            bbox_tuple = get_bbox_3d(
                current_colev,
                self.frame_col,
                self.posx,
                self.posy,
                self.posz,
                self.output_order,
            )
            bbox_tuple = fix_3d_convex_hull(
                df_tp, bbox_tuple[0], bbox_tuple[1], bbox_tuple[2], self.frame_col
            )
            self.viewer.add_surface(
                bbox_tuple,
                colormap="red",
                opacity=0.15,
                name=ARCOS_LAYERS["event_boundingbox"],
                shading="none",
            )

        current_step = list(self.viewer.dims.current_step)
        t_index = self.output_order.index("t")
        current_step[t_index] = frame
        self.viewer.dims.current_step = current_step


class BlitManager:
    """Class to manage blitting of the annotation. Copied from matplotlib docs."""

    def __init__(self, canvas, animated_artists=()):
        """
        Parameters
        ----------
        canvas (FigureCanvasAgg): The canvas to work with, this only works for
        sub-classes of the Agg
            canvas which have the `~FigureCanvasAgg.copy_from_bbox` and
            `~FigureCanvasAgg.restore_region` methods.

        animated_artists (Iterable[Artist]): List of the artists to manage
        """
        self.canvas = canvas
        self._bg = None
        self._artists = []

        for _artist in animated_artists:
            self.add_artist(_artist)
        # grab the background on every draw
        self.cid = canvas.mpl_connect("draw_event", self.on_draw)

    def on_draw(self, event):
        """Callback to register with 'draw_event'."""
        _canvas = self.canvas
        if event is not None:
            if event.canvas != _canvas:
                raise RuntimeError
        self._bg = _canvas.copy_from_bbox(_canvas.figure.bbox)
        self._draw_animated()

    def add_artist(self, art):
        """
        Add an artist to be managed.

        Parameters:
            art (Artist): The artist to be added.  Will be set to 'animated' (just
            to be safe).  *art* must be in the figure associated with
            the canvas this class is managing.

        """
        if art.figure != self.canvas.figure:
            raise RuntimeError
        art.set_animated(True)
        self._artists.append(art)

    def _draw_animated(self):
        """Draw all of the animated artists."""
        fig = self.canvas.figure
        for a in self._artists:
            fig.draw_artist(a)

    def update(self):
        """Update the screen with animated artists."""
        cv = self.canvas
        fig = cv.figure
        # paranoia in case we missed the draw event,
        if self._bg is None:
            self.on_draw(None)
        else:
            # restore the background
            cv.restore_region(self._bg)
            # draw all of the animated artists
            self._draw_animated()
            # update the GUI state
            cv.blit(fig.bbox)
        # let the GUI event loop process anything it has to do
        cv.flush_events()


class TimeSeriesPlots(QtWidgets.QWidget):
    """
    QWidget for plotting.
    Class to make a matplotlib figure canvas and add it to a Qwidget.
    Canvas, figure and axis objects can be acessed by self.canvas, self.fig and self.ax.
    This plots several different Timeseries plots such as Position/t plots,
    tracklength histogram and a measurment density plot.

    Attributes:
        parent (qtpy.QtWidgets.QWidget): Parent widget, optional
    """

    def __init__(self, parent=None):
        """Constructs class with given arguments
        Parameters:
            parent (qtpy.QtWidgets.QWidget): Parent widget, optional
        """
        super().__init__(parent=parent)

        # available plots
        self.plot_list = [
            "tracklength histogram",
            "measurment density plot",
            "measurment density plot rescaled",
            "original vs detreded",
            "x/t-plot",
            "y/t-plot",
        ]
        # will be set later in methods
        self._dataframe = pd.DataFrame()
        self._dataframe_resc = pd.DataFrame()
        self._frame_col = None
        self._track_id_col = None
        self._x_coord_col = None
        self._y_coord_col = None
        self._measurement = None
        self._measurement_resc_col = None
        self._object_id_number = None

        self._init_widgets()

    def _init_widgets(self):
        """
        Method to initialise a matplotlib figure canvas as well as a spinbox,
        Button and label widgets. Additionally, generates a
        matplotlib.backends.backend_qt5agg.FigureCanvas, a set plot style and axis,
        and populates it with a matplotlib.figure.Figure.
        These are the added to a QVboxlayout.
        """
        # creating spinbox widget
        self.sample_number = QtWidgets.QSpinBox()
        self.sample_number.setMinimum(1)
        self.sample_number.setMaximum(200)
        self.sample_number.setValue(20)

        self.button = QtWidgets.QPushButton("Update Plot")
        self.resc_check = QtWidgets.QCheckBox("Show Rescaled")
        self.resc_check.setVisible(False)
        self.resc_check.setChecked(True)
        self.orig_check = QtWidgets.QCheckBox("Show Original")
        self.orig_check.setVisible(False)
        self.orig_check.setChecked(True)

        # label
        self.spinbox_title = QtWidgets.QLabel("Sample Size")
        self.spinbox_title.setVisible(False)

        # creating a combo box widget
        self.combo_box = QtWidgets.QComboBox(self)
        self.combo_box.addItems(self.plot_list)

        # set up figure and axe objects
        with plt.style.context("dark_background"):
            plt.rcParams["axes.edgecolor"] = "#ffffff"
            self.fig = Figure(figsize=(2.5, 1.5))
            self.canvas = FigureCanvas(self.fig)
            self.ax = self.fig.add_subplot(111)

        self.toolbar = NavigationToolbar(self.canvas, self)

        # construct layout
        layout = QtWidgets.QVBoxLayout()
        layout_combobox = QtWidgets.QVBoxLayout()
        layout_spinbox = QtWidgets.QHBoxLayout()
        self.setLayout(layout)

        # add widgets to sub_layouts
        layout_combobox.addWidget(self.button)
        layout_combobox.addWidget(self.combo_box)

        layout_spinbox.addWidget(self.spinbox_title)
        layout_spinbox.addWidget(self.sample_number)
        layout_spinbox.addWidget(self.resc_check)
        layout_spinbox.addWidget(self.orig_check)

        layout.addLayout(layout_combobox)
        layout.addLayout(layout_spinbox)

        # add sublayouts together
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        self.setWindowTitle("Collective Events")
        self.combo_box.currentIndexChanged.connect(self._update)
        self._update_layout()
        self.button.clicked.connect(self._update_from_button)
        self.orig_check.stateChanged.connect(self._update)
        self.resc_check.stateChanged.connect(self._update)

    def _update_layout(self):
        self.fig.tight_layout(pad=0.1, w_pad=0.001, h_pad=0.05)

    def data_clear(self):
        """
        Method to clear the data from the plot.
        """
        self.ax.clear()
        self._dataframe = pd.DataFrame()
        self._dataframe_resc = pd.DataFrame()
        self._frame_col = None
        self._track_id_col = None
        self._x_coord_col = None
        self._y_coord_col = None
        self._measurement = None
        self._measurement_resc_col = None
        self._object_id_number = None

    def update_plot(
        self,
        dataframe: pd.DataFrame,
        dataframe_resc: pd.DataFrame,
        frame_col,
        track_id_col,
        x_coord_col,
        y_coord_col,
        measurement_col,
        measurement_resc_col,
        object_id_number=None,
    ):
        """
        Method to update the from the dropdown menu chosen
        matplotlibl plot with values from
        the stored_variables object dataframe.
        """
        self._dataframe = dataframe
        self._dataframe_resc = dataframe_resc
        self._frame_col = frame_col
        self._track_id_col = track_id_col
        self._x_coord_col = x_coord_col
        self._y_coord_col = y_coord_col
        self._measurement = measurement_col
        self._measurement_resc_col = measurement_resc_col
        self._object_id_number = object_id_number
        self._update()

    def _update_from_button(self):
        self._object_id_number = None
        self._update()

    def _update(self):
        # return plottype that should be plotted
        plottype = self.combo_box.currentText()
        # sample number for position/t-plots
        n_samples = self.sample_number.value()

        # check if some data was loaded already, otherwise do nothing
        if not self._dataframe.empty:
            self._init_empty_plot()

            # tracklength histogram
            if plottype == "tracklength histogram":
                self._tracklenght_histogram()

            # measurment density plot, kde
            elif plottype == "measurment density plot":
                self._meas_density_plot()

            elif plottype == "measurment density plot rescaled":
                self._meas_density_plot_rescaled()

            # xy/t plots
            elif plottype == "x/t-plot":
                self._xt_plot(n_samples)

            elif plottype == "y/t-plot":
                self._yt_plot(n_samples)

            elif plottype == "original vs detreded":
                self._orig_vs_detrended_plot()

            self.fig.canvas.draw_idle()
        else:
            show_info("No Data to plot")

    def _orig_vs_detrended_plot(self):
        self.sample_number.setVisible(False)
        self.spinbox_title.setVisible(False)

        self.resc_check.setVisible(True)
        self.orig_check.setVisible(True)

        if not self._dataframe_resc.empty and self._track_id_col:
            if self._object_id_number:
                vals = self._object_id_number
            else:
                vals = np.random.choice(
                    self._dataframe_resc[self._track_id_col].unique(),
                    1,
                    replace=False,
                )
                self._object_id_number = vals
            dataframe_resc_cp = (
                self._dataframe_resc.set_index(self._track_id_col)
                .loc[vals]
                .reset_index()
                .copy(deep=True)
            )
            grouped = dataframe_resc_cp.groupby(self._track_id_col)
            plot_data_types = []
            if self.resc_check.isChecked():
                plot_data_types.append(self._measurement_resc_col)
            if self.orig_check.isChecked():
                plot_data_types.append(self._measurement)
            for val in vals:
                df_g = grouped.get_group(val)
                if not plot_data_types:
                    self.ax.plot()
                else:
                    df_g.plot(
                        x=self._frame_col,
                        y=plot_data_types,
                        ax=self.ax,
                    )
                    x_val = df_g[df_g[f"{self._measurement}.bin"] != 0][self._frame_col]
                    y_val = np.repeat(self.ax.get_ylim()[0], x_val.size)
                    indices = np.where(np.diff(x_val) != 1)[0] + 1
                    x_split = np.split(x_val, indices)
                    y_split = np.split(y_val, indices)
                    for idx, (x_val, y_val) in enumerate(zip(x_split, y_split)):
                        if idx == 0:
                            self.ax.plot(x_val, y_val, color="red", lw=2, label="bin")
                        else:
                            self.ax.plot(x_val, y_val, color="red", lw=2)
            if plot_data_types:
                self.ax.legend(loc=2, prop={"size": 6})
            self.ax.set_xlabel("Frame")
            self.ax.set_ylabel("Mes Value")

    def _yt_plot(self, n_samples):
        self.sample_number.setVisible(True)
        self.spinbox_title.setVisible(True)
        if self._track_id_col:
            sample = pd.Series(self._dataframe[self._track_id_col].unique()).sample(
                n_samples, replace=True
            )
            pd_from_r_df = self._dataframe.loc[
                self._dataframe[self._track_id_col].isin(sample)
            ]
            df_grp = pd_from_r_df.groupby(self._track_id_col)
            for _, df in df_grp:
                self.ax.plot(df[self._frame_col], df[self._y_coord_col])
        self.ax.set_xlabel("Frame")
        self.ax.set_ylabel("Position Y")

    def _xt_plot(self, n_samples):
        self.sample_number.setVisible(True)
        self.spinbox_title.setVisible(True)
        self.resc_check.setVisible(False)
        self.orig_check.setVisible(False)
        if self._track_id_col:
            sample = pd.Series(self._dataframe[self._track_id_col].unique()).sample(
                n_samples, replace=True
            )
            pd_from_r_df = self._dataframe.loc[
                self._dataframe[self._track_id_col].isin(sample)
            ]
            df_grp = pd_from_r_df.groupby(self._track_id_col)
            for _, df in df_grp:
                self.ax.plot(df[self._frame_col], df[self._x_coord_col])
        self.ax.set_xlabel("Frame")
        self.ax.set_ylabel("Position X")

    def _meas_density_plot_rescaled(self):
        if not self._dataframe_resc.empty:
            measurement_resc_values = self._dataframe_resc[
                self._measurement_resc_col
            ].interpolate()
            if measurement_resc_values.size != 0:
                self.sample_number.setVisible(False)
                self.spinbox_title.setVisible(False)
                self.resc_check.setVisible(False)
                self.orig_check.setVisible(False)
                density = gaussian_kde(measurement_resc_values)
                x_val = np.linspace(
                    min(measurement_resc_values),
                    max(measurement_resc_values),
                    100,
                )
                y_val = density(x_val)
                self.ax.plot(x_val, y_val)
                self.ax.set_xlabel("measurement values")
                self.ax.set_ylabel("density")

    def _meas_density_plot(self):
        self.sample_number.setVisible(False)
        self.spinbox_title.setVisible(False)
        self.resc_check.setVisible(False)
        self.orig_check.setVisible(False)
        density = gaussian_kde(self._dataframe[self._measurement].interpolate())
        x_val = np.linspace(
            min(self._dataframe[self._measurement]),
            max(self._dataframe[self._measurement]),
            100,
        )
        y_val = density(x_val)
        self.ax.plot(x_val, y_val)
        self.ax.set_xlabel("measurement values")
        self.ax.set_ylabel("density")

    def _tracklenght_histogram(self):
        self.sample_number.setVisible(False)
        self.spinbox_title.setVisible(False)
        self.resc_check.setVisible(False)
        self.orig_check.setVisible(False)
        if self._track_id_col:
            track_length = self._dataframe.groupby(self._track_id_col).size()
            self.ax.hist(track_length)
        self.ax.set_xlabel("tracklength")
        self.ax.set_ylabel("counts")

    def _init_empty_plot(self):
        self.ax.cla()
        self.ax.spines["bottom"].set_color("white")
        self.ax.spines["top"].set_color("white")
        self.ax.spines["right"].set_color("white")
        self.ax.spines["left"].set_color("white")
        self.ax.xaxis.label.set_color("white")
        self.ax.yaxis.label.set_color("white")
        self.ax.tick_params(colors="white", which="both")
        self.ax.axis("on")



"""Main widget of the ARCOS GUI. Contains all other widgets.

Each widget is a tab in the main widget. Creates instances of all other widgets
as well as the data storage and layermaker.
The MainWidget is the entry point for the napari plugin.
"""
from __future__ import annotations

import weakref
from pathlib import Path
from typing import TYPE_CHECKING

from arcos_gui.layerutils import Layermaker
from arcos_gui.processing import DataStorage
from arcos_gui.widgets import (
    ArcosController,
    BottombarController,
    ExportController,
    FilterController,
    InputdataController,
    LayerpropertiesController,
    collevPlotWidget,
    tsPlotWidget,
)
from napari.utils.notifications import show_info
from qtpy import QtWidgets, uic

if TYPE_CHECKING:
    import napari.viewer


class _MainUI(QtWidgets.QWidget):
    UI_FILE = str(Path(__file__).parent / "_ui" / "main_widget.ui")

    # The UI_FILE above contains these objects:
    maintabwidget: QtWidgets.QTabWidget

    filter_groupBox: QtWidgets.QGroupBox
    inputdata_groupBox: QtWidgets.QGroupBox
    arcos_layout: QtWidgets.QVBoxLayout
    layer_prop_layout: QtWidgets.QVBoxLayout
    export_layout: QtWidgets.QVBoxLayout

    evplots_layout: QtWidgets.QVBoxLayout
    tsplots_layout: QtWidgets.QVBoxLayout

    bottom_bar_layout: QtWidgets.QVBoxLayout

    nbr_collev_display: QtWidgets.QLCDNumber
    help_button: QtWidgets.QPushButton

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setup_ui()

    def setup_ui(self):
        """Setup UI. Loads it from ui file."""
        uic.loadUi(self.UI_FILE, self)  # load QtDesigner .ui file


class MainWindow(QtWidgets.QWidget):
    """
    Widget allowing a user to import a csv file, filter this file,
    choose arcos parameters, choose LUT mappings aswell as shape sizes.
    """

    _instance = None

    def __init__(self, viewer: napari.viewer.Viewer, parent=None):
        """Constructs class with provided arguments."""
        super().__init__(parent=parent)
        self.viewer: napari.viewer.Viewer = viewer
        self._widget = _MainUI(self)
        MainWindow._instance = weakref.ref(self)

        self.setLayout(self._widget.layout())

        self.data = DataStorage()
        # self.data_storage_instance.set_verbose(
        #     True
        # )  # uncomment to make callbacks verbose

        self._input_controller = InputdataController(
            data_storage_instance=self.data,
            std_out=show_info,
            viewer=self.viewer,
            parent=self,
        )
        self._filter_controller = FilterController(
            viewer=self.viewer,
            data_storage_instance=self.data,
            parent=self,
        )
        self._arcos_widget = ArcosController(
            data_storage_instance=self.data,
            parent=self,
        )
        self._layer_prop_controller = LayerpropertiesController(
            viewer=self.viewer,
            data_storage_instance=self.data,
            parent=self,
        )
        self._ts_plots_widget = tsPlotWidget(self.viewer, self.data, self)
        self._collev_plots_widget = collevPlotWidget(self.viewer, self.data, self)
        self._export = ExportController(
            viewer=self.viewer,
            data_storage_instance=self.data,
            parent=self,
        )

        self._bottom_bar = BottombarController(
            data_storage_instance=self.data,
            parent=self._widget,
        )

        self._layermaker = Layermaker(self.viewer, self.data)
        self._add_widgets()
        self._connect_signals()

    @classmethod
    def get_last_instance(cls) -> MainWindow | None:
        """Returns the last instance of this class. Returns None if no instance exists."""
        instance = cls._instance() if callable(cls._instance) else None

        if instance:
            try:
                instance._widget.maintabwidget.currentIndex()
                return instance
            except (AttributeError, RuntimeError):
                # The Qt object is no longer valid
                del cls._instance

        return None

    def _connect_signals(self):
        self.data.arcos_binarization.value_changed.connect(
            self._layermaker.make_layers_bin
        )
        self.data.arcos_output.value_changed.connect(self._layermaker.make_layers_all)

    def _add_widgets(self):
        self._widget.inputdata_groupBox.layout().addWidget(
            self._input_controller.widget
        )
        self._widget.filter_groupBox.layout().addWidget(self._filter_controller.widget)
        self._widget.arcos_layout.addWidget(self._arcos_widget.widget)
        self._widget.layer_prop_layout.addWidget(self._layer_prop_controller.widget)
        self._widget.tsplots_layout.addWidget(self._ts_plots_widget)
        self._widget.evplots_layout.addWidget(self._collev_plots_widget)
        self._widget.export_layout.addWidget(self._export.widget)
        self._widget.bottom_bar_layout.addWidget(self._bottom_bar.widget)


if __name__ == "__main__":
    import napari

    viewer = napari.Viewer()
    win = MainWindow(viewer)
    dw = viewer.window.add_dock_widget(win, name="ARCOS", area="right")
    napari.run()



"""Widgets."""

__author__ = """Benjamin Graedel"""
__email__ = "benjamin.graedel@unibe.ch"
__version__ = "0.1.3"

from arcos_gui.widgets._arcos_widget import ArcosController
from arcos_gui.widgets._bottom_bar_widget import BottombarController
from arcos_gui.widgets._dialog_widgets import columnpicker
from arcos_gui.widgets._exporting_widget import ExportController
from arcos_gui.widgets._filter_widget import FilterController
from arcos_gui.widgets._input_data_widget import InputdataController
from arcos_gui.widgets._plot_widgets import collevPlotWidget, tsPlotWidget
from arcos_gui.widgets._visualization_settings_widget import LayerpropertiesController

__all__ = [
    "ArcosController",
    "ExportController",
    "FilterController",
    "InputdataController",
    "LayerpropertiesController",
    "tsPlotWidget",
    "collevPlotWidget",
    "columnpicker",
    "BottombarController",
]



"""Configuration file for the ARCOS GUI."""

import operator

from napari.utils import Colormap

OPERATOR_DICTIONARY = {
    "Divide": (operator.truediv, "Measurement_Ratio"),
    "Multiply": (operator.mul, "Measurement_Product"),
    "Add": (operator.add, "Measurement_Sum"),
    "Subtract": (operator.sub, "Measurement_Difference"),
}

COLOR_CYCLE = [
    "#1f77b4",
    "#aec7e8",
    "#ff7f0e",
    "#ffbb78",
    "#2ca02c",
    "#98df8a",
    "#d62728",
    "#ff9896",
    "#9467bd",
    "#c5b0d5",
    "#8c564b",
    "#c49c94",
    "#e377c2",
    "#f7b6d2",
    "#7f7f7f",
    "#c7c7c7",
    "#bcbd22",
    "#dbdb8d",
    "#17becf",
    "#9edae5",
]

TAB20 = Colormap(COLOR_CYCLE, "tab20", interpolation="zero")

ARCOS_LAYERS = {
    "all_cells": "All Cells",
    "active_cells": "Active Cells",
    "collective_events_cells": "Collective Events Cells",
    "event_hulls": "Collective Events",
    "event_boundingbox": "Event Bounding Box",
    "timestamp": "Timestamp",
}

AVAILABLE_OPTIONS_FOR_BATCH = [
    "arcos_output",
    "arcos_stats",
    "per_frame_statistics",
    "noodleplot",
    "statsplot",
]

ALLOWED_SETTINGS = [
    "file_name",
    "columns",
    "arcos_parameters",
    "min_max_meas",
    "point_size",
    "lut",
    "output_order",
    "min_max_tracklenght",
]


ARCOSPARAMETERS_DEFAULTS = {
    "interpolate_meas": False,
    "clip_meas": False,
    "clip_low": 0.0,
    "clip_high": 1.0,
    "smooth_k": 1,
    "bias_k": 5,
    "bias_method": "none",
    "polyDeg": 1,
    "bin_threshold": 0.5,
    "bin_peak_threshold": 0.5,
    "eps_method": "manual",
    "neighbourhood_size": 20.0,
    "epsPrev": 20.0,
    "min_clustersize": 5,
    "nprev": 1,
    "min_dur": 1,
    "total_event_size": 5,
    "add_convex_hull": True,
    "add_all_cells": True,
    "add_bin_cells": True,
    "bin_advanded_settings": False,
    "detect_advanced_options": False,
}



"""Collection of various functions for the GUI."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING

from napari.qt import get_stylesheet
from napari.settings import get_settings
from qtpy.QtCore import Qt, QTimer, Signal
from qtpy.QtGui import QValidator
from qtpy.QtWidgets import (
    QAction,
    QCheckBox,
    QDialog,
    QFileDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMenu,
    QMenuBar,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

if TYPE_CHECKING:
    import napari.viewer
    from qtpy import QtWidgets
    from superqt import QRangeSlider


class ThrottledCallback:
    def __init__(self, callback, max_interval):
        self.callback = callback
        self.max_interval = max_interval
        self.last_call_time = 0
        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.args, self.kwargs = None, None

    def __call__(self, *args, **kwargs):
        current_time = time.time()
        self.args, self.kwargs = args, kwargs  # store the latest args and kwargs

        if current_time - self.last_call_time > self.max_interval:
            self.last_call_time = current_time
            self.callback(*self.args, **self.kwargs)
        else:
            self.timer.stop()
            self.timer.timeout.connect(self._timeout_callback)
            self.timer.start(int(self.max_interval * 1000))

    def _timeout_callback(self):
        self.last_call_time = time.time()
        self.callback(*self.args, **self.kwargs)
        self.timer.timeout.disconnect(self._timeout_callback)


class OutputOrderValidator(QValidator):
    def __init__(self, vColsCore, parent=None):
        super().__init__(parent)
        self.vColsCore = vColsCore
        self.required_chars = ["t", "x", "y"]
        self.allowed_chars = ["t", "x", "y", "z"]
        self.max_length = 4

    def validate(self, string, pos):
        if len(string) > self.max_length:
            return QValidator.Invalid, string, pos

        for char in string:
            if char not in self.allowed_chars:
                return QValidator.Invalid, string, pos

        if len(set(string)) != len(string):
            return QValidator.Invalid, string, pos

        if len(string) < len(self.vColsCore):
            return QValidator.Intermediate, string, pos

        for char in self.required_chars:
            if char not in string:
                return QValidator.Intermediate, string, pos

        return QValidator.Acceptable, string, pos


class DirectoryInputDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Enter Directory Path")
        layout = QVBoxLayout(self)

        self.line_edit = QLineEdit(self)
        layout.addWidget(self.line_edit)

        button_layout = QHBoxLayout()
        self.ok_button = QPushButton("OK", self)
        self.ok_button.clicked.connect(self.accept)
        self.cancel_button = QPushButton("Cancel", self)
        self.cancel_button.clicked.connect(self.reject)
        button_layout.addWidget(self.ok_button)
        button_layout.addWidget(self.cancel_button)

        layout.addLayout(button_layout)


class MenuBarWidget(QMenuBar):
    go_to_directory = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.navigate_menu = QMenu("Navigate", self)

        # Add predefined directories as actions
        self.home_action = QAction("Home", self)
        self.home_action.triggered.connect(self._go_home)
        self.navigate_menu.addAction(self.home_action)

        # Add action to navigate to a custom directory
        self.custom_dir_action = QAction("Go to directory...", self)
        self.custom_dir_action.triggered.connect(self._open_directory_input)
        self.navigate_menu.addAction(self.custom_dir_action)

        self.addMenu(self.navigate_menu)

    def _go_home(self):
        # Signal or method to set the directory to home
        self.go_to_directory.emit(".")

    def _open_directory_input(self):
        dialog = DirectoryInputDialog(self)
        if dialog.exec_():
            path = dialog.line_edit.text()
            self.go_to_directory.emit(path)


class SelectionCheckboxWidget(QWidget):
    def __init__(self, selection_values, parent=None):
        super().__init__(parent)

        layout = QGridLayout(self)
        self.selection_title = QLabel("Select:")
        layout.addWidget(self.selection_title, 0, 0, 1, 4)

        self.selectAllCheckbox = QCheckBox("Select All")
        self.selectAllCheckbox.setChecked(True)
        layout.addWidget(self.selectAllCheckbox, 1, 0, 1, 4)

        self.checkboxes = {}
        for index, value in enumerate(selection_values):
            self.checkboxes[value] = QCheckBox(value.replace("_", " ").title())
            self.checkboxes[value].setChecked(True)
            row = 2 + index // 5
            col = index % 5
            layout.addWidget(self.checkboxes[value], row, col)

        # Connect "Select All" checkbox to selectAll function
        self.selectAllCheckbox.stateChanged.connect(self.selectAll)
        # Connect each individual checkbox to updateSelectAllState function
        for checkbox in self.checkboxes.values():
            checkbox.stateChanged.connect(self.updateSelectAllState)

        # Set stylesheet
        style_sheet = get_stylesheet(get_settings().appearance.theme)
        self.setStyleSheet(style_sheet)

        # Override mousePressEvent for selectAllCheckbox
        def new_mousePressEvent(
            event, original_method=self.selectAllCheckbox.mousePressEvent
        ):
            if self.selectAllCheckbox.checkState() == Qt.Unchecked:
                self.selectAllCheckbox.setCheckState(Qt.Checked)
            else:
                original_method(event)

        self.selectAllCheckbox.mousePressEvent = new_mousePressEvent

    def selectAll(self, state):
        if state == Qt.PartiallyChecked:
            return  # Ignore partially checked state

        isChecked = state == Qt.Checked

        # Block signals from individual checkboxes to prevent recursion
        for checkbox in self.checkboxes.values():
            checkbox.blockSignals(True)
            checkbox.setChecked(isChecked)
            checkbox.blockSignals(False)

        # Now, update the "Select All" checkbox without triggering its signal
        self.selectAllCheckbox.blockSignals(True)
        self.selectAllCheckbox.setCheckState(Qt.Checked if isChecked else Qt.Unchecked)
        self.selectAllCheckbox.blockSignals(False)

    def updateSelectAllState(self):
        checkedCount = sum(
            1 for checkbox in self.checkboxes.values() if checkbox.isChecked()
        )

        # Update the state of selectAllCheckbox without triggering its signal
        self.selectAllCheckbox.blockSignals(True)
        if checkedCount == 0:
            self.selectAllCheckbox.setCheckState(Qt.Unchecked)
        elif checkedCount == len(self.checkboxes):
            self.selectAllCheckbox.setCheckState(Qt.Checked)
        else:
            self.selectAllCheckbox.setCheckState(Qt.PartiallyChecked)
        self.selectAllCheckbox.blockSignals(False)


class SelectionDialog(QFileDialog):
    def __init__(self, selection_values, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setOption(QFileDialog.DontUseNativeDialog, True)

        # Menu bar
        self.menu_bar = MenuBarWidget(self)
        self.layout().setMenuBar(self.menu_bar)

        # Checkboxes
        self.selection_widget = SelectionCheckboxWidget(selection_values, self)
        self.layout().addWidget(self.selection_widget, 4, 0, 1, 4)

        # Connect signals
        self.menu_bar.go_to_directory.connect(self.setDirectory)

        style_sheet = get_stylesheet(get_settings().appearance.theme)
        self.setStyleSheet(style_sheet)

    def navigate_to_custom_directory(self):
        self.menu_bar._open_directory_input()

    def get_selected_options(self):
        """Returns a list of selected checkboxes' values excluding the 'Select All' checkbox."""
        return [
            value
            for value, checkbox in self.checkboxes.items()
            if checkbox.isChecked() and value != "Select All"
        ]


class BatchFileDialog(SelectionDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.checkboxes = self.selection_widget.checkboxes
        self.setFileMode(QFileDialog.Directory)
        self.setOption(QFileDialog.ShowDirsOnly, True)
        self.setWindowTitle("Select Folder to Batch Process")
        self.selection_widget.selection_title.setText("Select what to export:")


class ParameterFileDialog(SelectionDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.checkboxes = self.selection_widget.checkboxes
        self.setFileMode(QFileDialog.ExistingFile)
        self.setOption(QFileDialog.ShowDirsOnly, False)
        # filter for .yaml files
        self.setNameFilter("*.yaml")
        self.setWindowTitle("Select Parameter File")
        self.selection_widget.selection_title.setText("Select what to import:")


def remove_layers_after_columnpicker(viewer: napari.viewer.Viewer, arcos_layers: list):
    """Remove existing arcos layers before loading new data"""
    layer_list = get_layer_list(viewer)
    for layer in arcos_layers:
        if layer in layer_list:
            viewer.layers.remove(layer)


def get_layer_list(viewer: napari.viewer.Viewer):
    """Get list of open layers."""
    layer_list = [layer.name for layer in viewer.layers]
    return layer_list


def set_track_lenths(
    track_lenths_minmax: tuple,
    tracklenght_slider: QRangeSlider,
    min_tracklength_spinbox: QtWidgets.QDoubleSpinBox,
    max_tracklength_spinbox: QtWidgets.QDoubleSpinBox,
):
    """Set track length slider and spinboxes to min and max values of track lengths.

    Parameters
    ----------
    track_lenths_minmax : tuple
        Tuple of min and max track length.
    tracklenght_slider : QRangeSlider
    tracklenght_slider : QRangeSlider
    min_tracklength_spinbox : QtWidgets.QDoubleSpinBox
    max_tracklength_spinbox : QtWidgets.QDoubleSpinBox
    """
    if track_lenths_minmax[1] - track_lenths_minmax[0] > 1:
        tracklenght_slider.setMinimum(track_lenths_minmax[0])
        tracklenght_slider.setMaximum(track_lenths_minmax[1])

    min_tracklength_spinbox.setMinimum(track_lenths_minmax[0])
    max_tracklength_spinbox.setMinimum(track_lenths_minmax[0])

    min_tracklength_spinbox.setMaximum(track_lenths_minmax[1])
    max_tracklength_spinbox.setMaximum(track_lenths_minmax[1])

    min_tracklength_spinbox.setValue(track_lenths_minmax[0])
    max_tracklength_spinbox.setValue(track_lenths_minmax[1])



"""Popupdialog widgets for the arcos_gui package."""

from __future__ import annotations

from typing import Iterable, Union

from arcos_gui.processing import columnnames
from arcos_gui.tools import OPERATOR_DICTIONARY
from qtpy import QtWidgets
from qtpy.QtCore import Signal


class columnpicker(QtWidgets.QDialog):
    """Dialog to pick the columns of the data file.

    Parameters
    ----------
    parent : QtWidgets.QWidget, optional
        Parent widget, by default None
    columnames_instance : Union[columnnames, None], optional
        Instance of the columnnames class, by default None

    Attributes
    ----------
    frame : QtWidgets.QComboBox
        Combobox to select the frame column
    track_id : QtWidgets.QComboBox
        Combobox to select the track id column
    x : QtWidgets.QComboBox
        Combobox to select the x column
    y : QtWidgets.QComboBox
        Combobox to select the y column
    measurement : QtWidgets.QComboBox
        Combobox to select the measurement column
    measurement_math : QtWidgets.QComboBox
        Combobox to select the measurement math column
    second_measurement : QtWidgets.QComboBox
        Combobox to select the second measurement column
    ok_button : QtWidgets.QPushButton
        Button to confirm the selection
    abort_button : QtWidgets.QPushButton
        Button to abort the selection
    ok_pressed : bool
        True if ok button was pressed, False if abort button was pressed

    Signals
    -------
    aborted : Signal
        Signal emitted if abort button was pressed
    ok : Signal
        Signal emitted if ok button was pressed
    """

    aborted = Signal()
    ok = Signal()

    def __init__(
        self, parent=None, columnames_instance: Union[columnnames, None] = None
    ):
        super().__init__(parent)

        self._setupUi()
        self._add_tooltipps()
        self.columnames_instance = columnames_instance
        self.set_measurement_math(list(OPERATOR_DICTIONARY.keys()))
        self.measurement_math.setCurrentText("None")
        self.measurement_math.currentTextChanged.connect(
            self.toggle_visible_second_measurment
        )
        self.second_measurement.setVisible(False)
        self.label_7.setVisible(False)

        self.ok_button.clicked.connect(self._on_ok)
        self.abort_button.clicked.connect(self._on_abort)
        self.ok_pressed = False

    def _on_ok(self, event):
        _ = event
        self.ok_pressed = True
        self.accept()

    def _on_abort(self, event):
        _ = event
        self.ok_pressed = False
        self.reject()

    def _setupUi(self):
        self.setObjectName("columnpicker")
        self.grid_layout = QtWidgets.QGridLayout()
        self.grid_layout.setObjectName("gridLayout")
        self.frame = QtWidgets.QComboBox()
        self.frame.setObjectName("frame")

        self.label = QtWidgets.QLabel("Frame Column:")
        self.label.setObjectName("frame_label")

        self.track_id = QtWidgets.QComboBox()
        self.track_id.setObjectName("track_id")

        self.label_2 = QtWidgets.QLabel("Object id Column:")
        self.label_2.setObjectName("track_id_label")

        self.label_3 = QtWidgets.QLabel("X Coordinate Column:")
        self.label_3.setObjectName("x_coordinates_label")

        self.x_coordinates = QtWidgets.QComboBox()
        self.x_coordinates.setObjectName("x_coordinates")

        self.label_4 = QtWidgets.QLabel("Y Coordinate Column:")
        self.label_4.setObjectName("y_coordinates_label")

        self.y_coordinates = QtWidgets.QComboBox()
        self.y_coordinates.setObjectName("y_coordinates")

        self.label_5 = QtWidgets.QLabel("Z Coordinate Column:")
        self.label_5.setObjectName("z_coordinates_label")

        self.z_coordinates = QtWidgets.QComboBox()
        self.z_coordinates.setObjectName("z_coordinates")

        self.label_6 = QtWidgets.QLabel("Measurement Column:")
        self.label_6.setObjectName("measurment_label")

        self.measurement = QtWidgets.QComboBox()
        self.measurement.setObjectName("measurment")

        self.label_7 = QtWidgets.QLabel("Second Measurement Column:")
        self.label_7.setObjectName("second_measurment_label")

        self.second_measurement = QtWidgets.QComboBox()
        self.second_measurement.setObjectName("second_measurment")

        self.label_8 = QtWidgets.QLabel("Field of View/Position Column:")
        self.label_8.setObjectName("field_of_view_id_label")

        self.field_of_view_id = QtWidgets.QComboBox()
        self.field_of_view_id.setObjectName("field_of_view_id")

        self.label_9 = QtWidgets.QLabel("Additional Filter Column:")
        self.label_9.setObjectName("additional_filter_label")

        self.additional_filter = QtWidgets.QComboBox()
        self.additional_filter.setObjectName("additional_filter")

        self.label_10 = QtWidgets.QLabel("Math on first and second measurement:")
        self.label_10.setObjectName("measurement_math_label")

        self.measurement_math = QtWidgets.QComboBox()
        self.measurement_math.setObjectName("measurement_math")

        self.ok_button = QtWidgets.QPushButton("Ok")
        self.ok_button.setObjectName("Ok")

        self.abort_button = QtWidgets.QPushButton("Abort")
        self.abort_button.setObjectName("Abort")
        self.abort_button.setStyleSheet("background-color : #7C0A02; color : white")

        self.grid_layout.addWidget(self.frame, 1, 1, 1, 1)
        self.grid_layout.addWidget(self.label, 1, 0, 1, 1)
        self.grid_layout.addWidget(self.label_2, 2, 0, 1, 1)
        self.grid_layout.addWidget(self.track_id, 2, 1, 1, 1)
        self.grid_layout.addWidget(self.label_3, 3, 0, 1, 1)
        self.grid_layout.addWidget(self.x_coordinates, 3, 1, 1, 1)
        self.grid_layout.addWidget(self.label_4, 4, 0, 1, 1)
        self.grid_layout.addWidget(self.y_coordinates, 4, 1, 1, 1)
        self.grid_layout.addWidget(self.label_5, 5, 0, 1, 1)
        self.grid_layout.addWidget(self.z_coordinates, 5, 1, 1, 1)
        self.grid_layout.addWidget(self.label_6, 6, 0, 1, 1)
        self.grid_layout.addWidget(self.measurement, 6, 1, 1, 1)
        self.grid_layout.addWidget(self.label_7, 7, 0, 1, 1)
        self.grid_layout.addWidget(self.second_measurement, 7, 1, 1, 1)
        self.grid_layout.addWidget(self.label_8, 8, 0, 1, 1)
        self.grid_layout.addWidget(self.field_of_view_id, 8, 1, 1, 1)
        self.grid_layout.addWidget(self.label_9, 9, 0, 1, 1)
        self.grid_layout.addWidget(self.additional_filter, 9, 1, 1, 1)
        self.grid_layout.addWidget(self.label_10, 10, 0, 1, 1)
        self.grid_layout.addWidget(self.measurement_math, 10, 1, 1, 1)
        self.grid_layout.addWidget(self.ok_button, 12, 1, 1, 1)
        self.grid_layout.addWidget(self.abort_button, 12, 0, 1, 1)
        # add it to our layout
        self.setLayout(self.grid_layout)

    def _add_tooltipps(self):
        """Add tooltipps to the widgets."""
        self.frame.setToolTip("Select frame column in input data")
        self.track_id.setToolTip("Select object id column in input data")
        self.x_coordinates.setToolTip("Column with x coordinates")
        self.y_coordinates.setToolTip("Column with y coordinates")
        self.z_coordinates.setToolTip("Column with z coordinates")
        self.measurement.setToolTip("Column with measurement")
        self.second_measurement.setToolTip(
            "Select second measurement for measurement math"
        )
        self.field_of_view_id.setToolTip(
            "Select fov column in input data, select None if column does not exist"
        )
        self.additional_filter.setToolTip(
            "Select additional filter column, for example Well of a wellplate, select None if column does not exist"
        )
        self.measurement_math.setToolTip(
            "Choose operation to calculate the measurment to be used in\n\
            arcos calculation on first and second measurement"
        )

    def _add_item_data_pair(
        self, combobox: QtWidgets.QComboBox, text: Iterable, data: Iterable
    ):
        """Add items to comboboxes."""
        for i, j in zip(text, data):
            combobox.addItem(i, j)

    def set_column_names(self, column_names):
        """Set column names in comboboxes."""
        while "" in column_names:
            column_names.remove("")

        self.frame.clear()
        self.track_id.clear()
        self.x_coordinates.clear()
        self.y_coordinates.clear()
        self.z_coordinates.clear()
        self.measurement.clear()
        self.second_measurement.clear()
        self.field_of_view_id.clear()
        self.additional_filter.clear()

        self._add_item_data_pair(self.frame, column_names, column_names)
        self._add_item_data_pair(self.track_id, column_names, column_names)
        self._add_item_data_pair(self.x_coordinates, column_names, column_names)
        self._add_item_data_pair(self.y_coordinates, column_names, column_names)
        self._add_item_data_pair(self.z_coordinates, column_names, column_names)
        self._add_item_data_pair(self.measurement, column_names, column_names)
        self._add_item_data_pair(self.second_measurement, column_names, column_names)
        self._add_item_data_pair(self.field_of_view_id, column_names, column_names)
        self._add_item_data_pair(self.additional_filter, column_names, column_names)

        self.additional_filter.addItem("None", None)
        self.field_of_view_id.addItem("None", None)
        self.z_coordinates.addItem("None", None)
        self.second_measurement.addItem("None", None)
        self.track_id.addItem("None", None)

        self.additional_filter.setCurrentText("None")
        self.field_of_view_id.setCurrentText("None")
        self.z_coordinates.setCurrentText("None")
        self.second_measurement.setCurrentText("None")
        self.track_id.setCurrentText("None")

    def set_measurement_math(self, measurement_math):
        """Set the measurement math options in the dialog."""
        self.measurement_math.clear()
        self._add_item_data_pair(
            self.measurement_math, measurement_math, measurement_math
        )
        self.measurement_math.addItem("None", None)

    def toggle_visible_second_measurment(self):
        """Toggles visibility of second measurement column."""
        curr_value = self.measurement_math.currentText()
        if curr_value in ["None", "1/X"]:
            self.second_measurement.setVisible(False)
            self.label_7.setVisible(False)
        else:
            self.second_measurement.setVisible(True)
            self.label_7.setVisible(True)

    @property
    def get_column_names(self):
        """Returns a tuple of all column names in the columnnames dialog."""
        return [
            self.frame.currentData(),
            self.track_id.currentData(),
            self.x_coordinates.currentData(),
            self.y_coordinates.currentData(),
            self.z_coordinates.currentData(),
            self.measurement.currentData(),
            self.second_measurement.currentData(),
            self.field_of_view_id.currentData(),
            self.additional_filter.currentData(),
            self.measurement_math.currentData(),
        ]

    @property
    def settable_columns(self):
        """Returns a tuple of all settable columns in the columnnames dialog."""
        return (
            self.frame,
            self.track_id,
            self.x_coordinates,
            self.y_coordinates,
            self.z_coordinates,
            self.measurement,
            self.second_measurement,
            self.field_of_view_id,
            self.additional_filter,
            self.measurement_math,
        )

    @property
    def as_columnames_object(self):
        """Retunrs a new columnames object with the current column names."""
        # set column_names
        columnames_instance = columnnames()
        columnames_instance.frame_column = self.frame.currentData()
        columnames_instance.position_id = self.field_of_view_id.currentData()
        columnames_instance.object_id = self.track_id.currentData()
        columnames_instance.x_column = self.x_coordinates.currentData()
        columnames_instance.y_column = self.y_coordinates.currentData()
        columnames_instance.z_column = self.z_coordinates.currentData()
        columnames_instance.measurement_column_1 = self.measurement.currentData()
        columnames_instance.measurement_column_2 = self.second_measurement.currentData()
        columnames_instance.additional_filter_column = (
            self.additional_filter.currentData()
        )
        columnames_instance.measurement_math_operation = (
            self.measurement_math.currentData()
        )

        return columnames_instance


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication(sys.argv)
    window = columnpicker()
    window.set_column_names(["a", "b", "c", "d", "e", "f", "g", "h", "i"])
    window.show()
    print(window.get_column_names)
    sys.exit(app.exec_())



from __future__ import annotations

from typing import Literal

import napari
import pandas as pd
from arcos_gui import _main_widget
from arcos_gui.processing import columnnames
from arcos_gui.sample_data import load_real_dataset, load_synthetic_dataset


def open_plugin(viewer: napari.Viewer):
    """Main function. Adds plugin dock widget to napari viewer.

    Parameters
    ----------
    viewer : napari.Viewer
        Napari viewer to add the plugin to.

    Returns
    -------
    plugin : _main_widget.MainWindow
        The plugin instance.
    """
    plugin: _main_widget.MainWindow
    viewer, plugin = viewer.window.add_plugin_dock_widget(
        plugin_name="arcos-gui", widget_name="ARCOS Main Widget"
    )
    return plugin


def load_sample_data(
    sample_data: Literal["synthetic", "real"] = "synthetic",
    plugin: _main_widget.MainWindow | None = None,
):
    """Load sample data into arcos-gui.

    Parameters
    ----------
    plugin : _main_widget.MainWindow | None
        If None, will try to get the last instance of the plugin (can fail if the plugin has been closed before
        and some stray reference is hanging arround).
        If not None, will use the given plugin instance. Plugin can be opened with open_plugin(viewer).
    sample_data : Literal['synthetic', 'real']
        Which sample data to load. Either synthetic or real.
    """
    if sample_data == "synthetic":
        load_synthetic_dataset(plugin=plugin)
    elif sample_data == "real":
        load_real_dataset(plugin=plugin)
    else:
        raise ValueError("sample_data must be either 'synthetic' or 'real'")


def load_dataframe(
    df: pd.DataFrame,
    frame_column: str,
    track_id_column: str | None,
    x_column: str,
    y_column: str,
    z_column: str | None,
    measurement_column: str,
    measurement_column_2: str | None,
    fov_column: str | None,
    additional_filter_column: str | None,
    measurement_math_operation: Literal[
        "Divide", "Multiply", "Add", "Subtract", None
    ] = None,
    plugin: _main_widget.MainWindow | None = None,
):
    """Load dataframe into arcos-gui. Bypasses the file loading dialog and columnpicker dialog.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to load, in a long table format where each row defines the object's
        location, time, and the measurement value.
    frame_column : str
        Name of the column in the dataframe that defines the timepoint of the measurement.
    track_id_column : str | None
        Name of the column in the dataframe that defines the track id of the object.
        If None, some functions (detrending, plotting, statistics) will not be available.
    x_column : str
        Name of the column in the dataframe that defines the x coordinate of the objects.
    y_column : str
        Name of the column in the dataframe that defines the y coordinate of the objects.
    z_column : str | None
        Name of the column in the dataframe that defines the z coordinate of the objects.
        If None, the data will be treated as 2D.
    fov_column : str | None
        Name of the column in the dataframe that defines the field of view of the objects, only
        used to filter the data.
        If None, the data will be treated as a single field of view.
    measurement_column : str
        Name of the column in the dataframe that defines the measurement value.
    measurement_column_2 : str | None
        Name of the column in the dataframe that defines the second measurement value.
        The second measurement is only used when performing a measurement math operation. Such as
        dividing the measurement by the second measurement.
    additional_filter_column : str | None
        Name of the column in the dataframe that defines the additional filter. Can be used to filter
        the data by e.g. Well ID.
    measurement_math_operation : Literal['Divide', "Multiply", "Add", "Subtract", None]
        Math operation to perform on the measurements. If None, no operation will be performed. And the
        second measurement column will be ignored.
    plugin : _main_widget.MainWindow | None
        If None, will try to get the last instance of the plugin (can fail if the plugin has been closed before
        and some stray reference is hanging arround).
        If not None, will use the given plugin instance. Plugin can be opened with open_plugin(viewer).
    """
    columns = columnnames(
        frame_column=frame_column,
        x_column=x_column,
        y_column=y_column,
        z_column=z_column,
        object_id=track_id_column,
        position_id=fov_column,
        measurement_column_1=measurement_column,
        measurement_column_2=measurement_column_2,
        additional_filter_column=additional_filter_column,
        measurement_math_operation=measurement_math_operation,
    )
    if plugin is None:
        plugin = _main_widget.MainWindow.get_last_instance()
    if not plugin:
        _plugin = open_plugin(napari.current_viewer())
    else:
        _plugin = plugin

    try:
        _plugin._input_controller.load_from_dataframe(dataframe=df, columns=columns)
    except RuntimeError:
        print("Cannot find the plugin. Opening a new one.")
        _plugin = open_plugin(napari.current_viewer())
        _plugin._input_controller.load_from_dataframe(dataframe=df, columns=columns)
    _plugin._widget.maintabwidget.setCurrentIndex(1)


def load_dataframe_with_columnpicker(df, plugin: _main_widget.MainWindow | None = None):
    """Load dataframe into arcos-gui. Opens a columnpicker dialog to select the columns.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to load, in a long table format where each row defines the object's
        location, time, and the measurement value.
        plugin : _main_widget.MainWindow | None
        If None, will try to get the last instance of the plugin (can fail if the plugin has been closed before and
        some stray reference is hanging arround).
        If not None, will use the given plugin instance. Plugin can be opened with open_plugin(viewer).
    """
    plugin = _main_widget.MainWindow.get_last_instance()
    if not plugin:
        _plugin = open_plugin(napari.current_viewer())
    else:
        _plugin = plugin

    try:
        _plugin._input_controller.load_from_dataframe(df, columns=None)
    except RuntimeError:
        print("Cannot find the plugin. Opening a new one.")
        _plugin = open_plugin(napari.current_viewer())
        _plugin._input_controller.load_from_dataframe_with_columnpicker(df)


def filter_data(
    track_length: tuple[int, int] | None = None,
    fov_id: str | None = None,
    additional_filter: str | None = None,
    plugin: _main_widget.MainWindow | None = None,
):
    """Filter data by track length and track id.

    Parameters
    ----------
    track_length : tuple[int, int] | None
        Tuple of minimum and maximum track length.
    fov_id : str | None
        Field of view id to filter by. If None,
        assumes that the data is from a single field of view.
    additional_filter : str | None
        Additional filter to filter by. If None,
        assumes that the data is not filtered by an additional filter.
    plugin : _main_widget.MainWindow | None
        If None, will try to get the last instance of the plugin (can fail if the plugin has been closed before
        and some stray reference is hanging arround).
        If not None, will use the given plugin instance. Plugin can be opened with open_plugin(viewer).

    """
    if plugin is None:
        plugin = _main_widget.MainWindow.get_last_instance()
    if not plugin:
        _plugin = open_plugin(napari.current_viewer())
    else:
        _plugin = plugin

    try:
        if fov_id is not None:
            _plugin._filter_controller.widget.position.setCurrentText(fov_id)
        if additional_filter is not None:
            _plugin._filter_controller.widget.additional_filter_combobox.setCurrentText(
                additional_filter
            )
        if track_length is not None:
            _plugin._filter_controller.widget.min_tracklength_spinbox.setValue(
                track_length[0]
            )
            _plugin._filter_controller.widget.max_tracklength_spinbox.setValue(
                track_length[1]
            )
    except RuntimeError:
        raise RuntimeError(
            "Cannot find the plugin. Either specify a plugin or open one."
        )

    _plugin._filter_controller.widget.filter_input_data.click()


def run_binarization_only(
    bias_method: Literal["none", "runmed", "lm"],
    smooth_k: int,
    bias_k: int,
    polyDeg: int,
    bin_peak_threshold: float,
    bin_threshold: float,
    interpolate: bool = True,
    clip: bool = False,
    clip_range: tuple[float, float] = (0.01, 0.99),
    plugin: _main_widget.MainWindow | None = None,
):
    """Run binarization only. Bypasses the binarization dialog.

    Parameters
    ----------
    bias_method : Literal['none', 'runmed', 'lm']
        Method to use for detrending.
    smooth_k : int
        Size of the local short-range smoothing kernel.
    bias_k : int
        Size of the global long-range smoothing kernel.
    polyDeg : int
        Degree of the polynomial to use for bias correction. Used only when bias_method is 'lm'.
    bin_peak_threshold : float
        Threshold above which trajectories are rescaled.
    bin_threshold : float
        Threshold to use for binarization.
    interpolate : bool
        Whether to interpolate the data before binarization.
    clip : bool
        Whether to clip the data before binarization.
    clip_range : tuple[float, float]
        Range to clip the data to. Used only when clip is True.
    plugin : _main_widget.MainWindow | None
        If None, will try to get the last instance of the plugin (can fail if the plugin has been closed before and
        some stray reference is hanging arround).
        If not None, will use the given plugin instance. Plugin can be opened with open_plugin(viewer).
    """
    if plugin is None:
        plugin = _main_widget.MainWindow.get_last_instance()
    if not plugin:
        _plugin = open_plugin(napari.current_viewer())
    else:
        _plugin = plugin

    try:
        _plugin._arcos_widget.widget.bias_method.setCurrentText(bias_method)
        _plugin._arcos_widget.widget.smooth_k.setValue(smooth_k)
        _plugin._arcos_widget.widget.bias_k.setValue(bias_k)
        _plugin._arcos_widget.widget.polyDeg.setValue(polyDeg)
        _plugin._arcos_widget.widget.bin_peak_threshold.setValue(bin_peak_threshold)
        _plugin._arcos_widget.widget.bin_threshold.setValue(bin_threshold)
        _plugin._arcos_widget.widget.interpolate_meas.setChecked(interpolate)
        _plugin._arcos_widget.widget.clip_meas.setChecked(clip)
        _plugin._arcos_widget.widget.clip_low.setValue(clip_range[0])
        _plugin._arcos_widget.widget.clip_high.setValue(clip_range[1])

        _plugin._arcos_widget.widget.run_binarization_only.click()
    except RuntimeError:
        raise RuntimeError(
            "Cannot find the plugin. Either specify a plugin or open one."
        )


def run_arcos(
    bias_method: Literal["none", "runmed", "lm"],
    smooth_k: int,
    bias_k: int,
    polyDeg: int,
    bin_peak_threshold: float,
    bin_threshold: float,
    interpolate: bool = True,
    clip: bool = False,
    clip_range: tuple[float, float] = (0.01, 0.99),
    eps_estimation: Literal["median", "kneepoint", "manual"] = "manual",
    eps: float = 0.1,
    eps_prev: float | None = None,
    min_clustersize: int = 3,
    min_duration: int = 3,
    min_event_size: int = 3,
    add_convex_hull: bool = True,
    plugin: _main_widget.MainWindow | None = None,
):
    """Run ARCoS. Bypasses the ARCoS dialog.

    Parameters
    ----------
    bias_method : Literal['none', 'runmed', 'lm']
        Method to use for detrending.
    smooth_k : int
        Size of the local short-range smoothing kernel.
    bias_k : int
        Size of the global long-range smoothing kernel.
    polyDeg : int
        Degree of the polynomial to use for bias correction. Used only when bias_method is 'lm'.
    bin_peak_threshold : float
        Threshold above which trajectories are rescaled.
    bin_threshold : float
        Threshold to use for binarization.
    interpolate : bool
        Whether to interpolate the data before binarization.
    clip : bool
        Whether to clip the data before binarization.
    clip_range : tuple[float, float]
        Range to clip the data to. Used only when clip is True.
    eps_estimation : Literal['median', 'kneepoint', 'manual']
        Method to use for estimating the epsilon parameter for DBSCAN.
    eps : float
        Epsilon parameter for DBSCAN. Used only when eps_estimation is 'manual'.
    eps_prev : float
        Distance to link clusters between frames. if None, uses eps.
    min_clustersize : int
        Minimum number of points in a cluster.
    min_duration : int
        Minimum number of frames for a cluster to be considered an event. Otherwise,
        it is filtered out.
    min_event_size : int
        Minimum number of points in an event. Otherwise, it is filtered out.
        Works only if a track_id column is present in the data.
    add_convex_hull : bool
        Whether to add a convex hull around the events.
    plugin : _main_widget.MainWindow | None
        If None, will try to get the last instance of the plugin (can fail if the plugin has been closed before and
        some stray reference is hanging arround).
        If not None, will use the given plugin instance. Plugin can be opened with open_plugin(viewer).
    """
    if plugin is None:
        plugin = _main_widget.MainWindow.get_last_instance()
    if not plugin:
        _plugin = open_plugin(napari.current_viewer())
    else:
        _plugin = plugin

    try:
        _plugin._arcos_widget._update_what_to_run_all()
    except RuntimeError:
        raise RuntimeError(
            "Cannot find the plugin. Either specify a plugin or open one."
        )
    _plugin._arcos_widget.widget.bias_method.setCurrentText(bias_method)
    _plugin._arcos_widget.widget.smooth_k.setValue(smooth_k)
    _plugin._arcos_widget.widget.bias_k.setValue(bias_k)
    _plugin._arcos_widget.widget.polyDeg.setValue(polyDeg)
    _plugin._arcos_widget.widget.bin_peak_threshold.setValue(bin_peak_threshold)
    _plugin._arcos_widget.widget.bin_threshold.setValue(bin_threshold)
    _plugin._arcos_widget.widget.interpolate_meas.setChecked(interpolate)
    _plugin._arcos_widget.widget.clip_meas.setChecked(clip)
    _plugin._arcos_widget.widget.clip_low.setValue(clip_range[0])
    _plugin._arcos_widget.widget.clip_high.setValue(clip_range[1])
    _plugin._arcos_widget.widget.eps_estimation_combobox.setCurrentText(eps_estimation)
    _plugin._arcos_widget.widget.neighbourhood_size.setValue(eps)
    if eps_prev is not None:
        _plugin._arcos_widget.widget.Cluster_linking_dist_checkbox.setChecked(True)
        _plugin._arcos_widget.widget.epsPrev_spinbox.setValue(eps_prev)
    else:
        _plugin._arcos_widget.widget.Cluster_linking_dist_checkbox.setChecked(False)

    _plugin._arcos_widget.widget.min_clustersize.setValue(min_clustersize)
    _plugin._arcos_widget.widget.min_dur.setValue(min_duration)
    _plugin._arcos_widget.widget.total_event_size.setValue(min_event_size)
    _plugin._arcos_widget.widget.add_convex_hull_checkbox.setChecked(add_convex_hull)

    _plugin._arcos_widget.widget.update_arcos.click()


def get_arcos_output(plugin: _main_widget.MainWindow | None = None):
    """Get the ARCoS output dataframe.

    Parameters
    ----------
    plugin : _main_widget.MainWindow | None
        If None, will try to get the last instance of the plugin (can fail if the plugin has been closed before and
        some stray reference is hanging arround).
        If not None, will use the given plugin instance. Plugin can be opened with open_plugin(viewer).

    Returns
    -------
    df : pd.DataFrame
        Dataframe with the ARCoS output.
    stats : pd.DataFrame
        Dataframe with summary statistics of the ARCOS output.
    """
    if plugin is None:
        plugin = _main_widget.MainWindow.get_last_instance()
    if not plugin:
        _plugin = open_plugin(napari.current_viewer())
    else:
        _plugin = plugin

    try:
        return _plugin.data.arcos_output.value, _plugin.data.arcos_stats.value
    except RuntimeError:
        raise RuntimeError(
            "Cannot find the plugin. Either specify a plugin or open one."
        )


def get_current_arcos_plugin():
    """Get the current ARCOS plugin instance.

    Returns
    -------
    plugin : _main_widget.MainWindow
        The plugin instance.
    """
    plugin = _main_widget.MainWindow.get_last_instance()
    if not plugin:
        plugin = open_plugin(napari.current_viewer())
    return plugin



"""Utility functoins to simplify layer creation."""

__author__ = """Benjamin Graedel"""
__email__ = "benjamin.graedel@unibe.ch"
__version__ = "0.1.3"

from arcos_gui.layerutils._layer_maker import Layermaker

__all__ = ["Layermaker"]



"""This module contains the InputDataWidget class.

This widget allows the user to import a csv file and choose the columns to use."""

from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Callable

import pandas as pd
from arcos_gui.processing import (
    DataFrameMatcher,
    DataLoader,
    preprocess_data,
    read_data_header,
)
from arcos_gui.widgets._dialog_widgets import columnpicker
from napari import viewer
from napari.layers import Labels, Tracks
from qtpy import QtCore, QtWidgets, uic
from qtpy.QtCore import Signal
from qtpy.QtGui import QIcon, QMovie

if TYPE_CHECKING:
    from arcos_gui.processing import DataStorage, columnnames

# icons
ICONS = Path(__file__).parent.parent / "_icons"


class _input_dataUI(QtWidgets.QWidget):
    filename_changed = Signal(str)
    closing = Signal()
    last_path = None

    UI_FILE = str(Path(__file__).parent.parent / "_ui" / "Input_data.ui")

    from_csv_selector: QtWidgets.QRadioButton
    from_layers_selector: QtWidgets.QRadioButton

    file_LineEdit: QtWidgets.QLineEdit
    load_data_button: QtWidgets.QPushButton
    browse_file: QtWidgets.QPushButton
    loading: QtWidgets.QLabel

    data_layer_selector_label: QtWidgets.QLabel
    data_layer_selector: QtWidgets.QListWidget

    tracks_layer_selector_label: QtWidgets.QLabel
    tracks_layer_selector: QtWidgets.QComboBox

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setup_ui()

    def setup_ui(self):
        uic.loadUi(self.UI_FILE, self)  # load QtDesigner .ui file
        # set text of Line edit
        self.browse_file.clicked.connect(self._browse_files)
        # set up file browser
        self.file_LineEdit.setText(".")
        # set icons
        browse_file_icon = QIcon(str(ICONS / "folder-open-line.svg"))
        self.loading_icon = QMovie(str(ICONS / "Dual Ring-1s-200px.gif"))
        self.loading_icon.setScaledSize(QtCore.QSize(40, 40))
        # self.loading.setMovie(self.loading_icon)
        self.browse_file.setIcon(browse_file_icon)
        self.load_data_button.setIcon(QIcon(self.loading_icon.currentPixmap()))
        self.loading_icon.stop()

        # set up list widget
        self.data_layer_selector.setSelectionMode(
            QtWidgets.QAbstractItemView.ExtendedSelection
        )

        # Connect radio buttons to slots
        self.from_csv_selector.toggled.connect(self.toggle_csv_selection_widgets)
        self.from_layers_selector.toggled.connect(self.toggle_layer_selection_widgets)

        # Initialize the widgets based on the selected option
        self.toggle_csv_selection_widgets(self.from_csv_selector.isChecked())
        self.toggle_layer_selection_widgets(self.from_layers_selector.isChecked())

    def toggle_csv_selection_widgets(self, toggled):
        # Show or hide CSV-related widgets based on the toggled state
        self.file_LineEdit.setVisible(toggled)
        self.browse_file.setVisible(toggled)

        self.file_LineEdit.setEnabled(toggled)
        self.browse_file.setEnabled(toggled)

    def toggle_layer_selection_widgets(self, toggled):
        # Show or hide Layers-related widgets based on the toggled state
        self.data_layer_selector_label.setVisible(toggled)
        self.data_layer_selector.setVisible(toggled)

        self.data_layer_selector_label.setEnabled(toggled)
        self.data_layer_selector.setEnabled(toggled)

        self.tracks_layer_selector_label.setVisible(toggled)
        self.tracks_layer_selector.setVisible(toggled)

    def _browse_files(self):
        """Opens a filedialog and saves path as a string in self.filename"""
        if self.last_path is None:
            self.last_path = str(Path.home())
        filename = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "Load CSV file",
            str(self.last_path),
            "csv(*.csv);; csv.gz(*.csv.gz);;",
        )
        self.last_path = str(Path(filename[0]).parent)
        if filename[0] == "":
            return
        self.file_LineEdit.setText(filename[0])
        self.filename_changed.emit(filename[0])

    def _set_loading_icon(self, frame=None):
        self.load_data_button.setIcon(QIcon(self.loading_icon.currentPixmap()))

    def _hide_loading_icon(self):
        self.load_data_button.setIcon(QIcon())

    def start_loading_icon(self):
        """Start loading icon animation."""
        self.loading_icon.start()
        self.loading_icon.frameChanged.connect(self._set_loading_icon)

    def stop_loading_icon(self):
        """Stop loading icon animation."""
        self.loading_icon.stop()
        self._hide_loading_icon()

    def closeEvent(self, event):
        self.closing.emit()
        event.accept()


class InputdataController:
    """Widget to import a csv file and choose the columns to use."""

    def __init__(
        self,
        data_storage_instance: DataStorage,
        std_out: Callable,
        viewer: viewer,
        parent=None,
    ):
        self.widget = _input_dataUI(parent)
        self.picker = columnpicker(self.widget)
        self.viewer = viewer

        self.data_storage_instance = data_storage_instance
        self.std_out = std_out

        self._connect_signals()
        self._update_labels_layers_list()
        self._update_tracks_layers_list()

    def _connect_signals(self):
        """Connects signals and slots."""
        self.widget.filename_changed.connect(self._update_filename)
        self.widget.load_data_button.clicked.connect(self._load_data)
        self.widget.closing.connect(self.closeEvent)
        self.viewer.layers.selection.events.changed.connect(self._on_selection)
        self.data_storage_instance.file_name.value_changed.connect(
            self._update_filename_from_datastorage
        )

    def _on_selection(self, event=None):
        num_labels_in_viewer = len(
            [label for label in self.viewer.layers if isinstance(label, Labels)]
        )
        if num_labels_in_viewer != self.widget.data_layer_selector.size():
            self._update_labels_layers_list()
            self._update_tracks_layers_list()

    def _update_filename(self, filename):
        """Updates the filename in the data storage instance."""
        self.data_storage_instance.file_name.value = filename

    def _update_filename_from_datastorage(self):
        """Updates the filename in the widget from the data storage instance."""
        self.data_storage_instance.file_name.toggle_callback_block(True)
        self.widget.file_LineEdit.setText(self.data_storage_instance.file_name.value)
        self.data_storage_instance.file_name.toggle_callback_block(False)

    def load_sample_data(self, path, columns):
        """Loads sample data from a given path and sets the column names.

        Parameters
        ----------
        path : str
            Path to the sample data file.
        columns : columnames instance
            Instance of the columnames class.
        """
        self.widget.from_csv_selector.setChecked(True)
        self.widget.file_LineEdit.setText(path)
        self.data_storage_instance.columns.value = columns
        self.widget.load_data_button.click()
        self.data_storage_instance.file_name.value = path

    def load_from_dataframe(self, dataframe: pd.DataFrame, columns: columnnames | None):
        """Loads data from a dataframe into the plugin.

        Parameters
        ----------
        dataframe : pandas.DataFrame
            Dataframe to load.
        columns : columnnames instance, optional
            Instance of the columnames class. If None the columnpicker will be opened.
        """
        if columns is None:
            self._open_columnpicker_from_dataframe(dataframe)
            return

        self._set_datastorage_to_default()
        self.data_storage_instance.columns.value = columns

        out_meas_name, dataframe_processed = preprocess_data(
            dataframe, self.data_storage_instance.columns.value
        )
        # set new measurmement name and data
        self.data_storage_instance.columns.value.measurement_column = out_meas_name
        self.data_storage_instance.original_data.value = dataframe_processed

    def _load_data(self):
        if self.widget.from_csv_selector.isChecked():
            self._open_columnpicker_from_csv()
        elif self.widget.from_layers_selector.isChecked():
            self._open_from_layers()
        else:
            raise ValueError("No data source selected")

    def _open_columnpicker_from_csv(self):
        """Opens a columnpicker window."""
        extension = [".csv", ".csv.gz"]
        csv_file = self.widget.file_LineEdit.text()
        if not csv_file.endswith(tuple(extension)):
            self.std_out("File type not supported")
            return
        if not Path(csv_file).exists():
            self.std_out("File does not exist")
            return
        csv_file = self.widget.file_LineEdit.text()
        columns, delimiter_value = read_data_header(csv_file)
        old_picked_columns = (
            self.data_storage_instance.columns.value.pickablepickable_columns_names
        )
        self.picker = columnpicker(
            parent=self.widget,
            columnames_instance=self.data_storage_instance.columns.value,
        )

        self.picker.set_column_names(columns)
        self._set_choices_names_from_previous(self.picker, old_picked_columns)
        self.picker.show()
        self._run_data_loading(csv_file, delimiter_value)

    def _open_columnpicker_from_dataframe(self, df: pd.DataFrame):
        self.picker = columnpicker(
            parent=self.widget,
            columnames_instance=self.data_storage_instance.columns.value,
        )

        self.picker.accepted.connect(partial(self._succesfully_loaded, df))

        self.picker.set_column_names(df.columns)
        old_picked_columns = (
            self.data_storage_instance.columns.value.pickablepickable_columns_names
        )
        self._set_choices_names_from_previous(self.picker, old_picked_columns)
        self.picker.show()

    def _open_columnpicker_from_layers(self, df: pd.DataFrame):
        self.picker = columnpicker(
            parent=self.widget,
            columnames_instance=self.data_storage_instance.columns.value,
        )

        self.picker.accepted.connect(partial(self._succesfully_loaded_from_layer, df))
        self.picker.set_column_names(df.columns)

        if self.widget.tracks_layer_selector.currentText() != "None":
            self.picker.track_id.addItem(
                "From napari tracks layer", "From napari tracks layer"
            )
            self.picker.track_id.setCurrentText("From napari tracks layer")

        old_picked_columns = (
            self.data_storage_instance.columns.value.pickablepickable_columns_names
        )
        self._set_choices_names_from_previous(self.picker, old_picked_columns)
        self.picker.show()

    def _open_from_layers(self):
        df = self._convert_selected_layer_properties_to_dataframe()
        if df is None:
            return
        self._open_columnpicker_from_layers(df)

    def _update_labels_layers_list(self):
        selected_labels = self._get_selected_labels_layers()
        self._available_labels = []
        self.widget.data_layer_selector.clear()
        for layer in self.viewer.layers:
            if isinstance(layer, (Labels)):
                item = QtWidgets.QListWidgetItem(layer.name)
                self._available_labels.append(layer)
                self.widget.data_layer_selector.addItem(item)
                if layer in selected_labels:
                    item.setSelected(True)

    def _update_tracks_layers_list(self):
        selected_tracks = self._get_selected_tracks_layers()
        self._available_tracks = ["None"]
        self.widget.tracks_layer_selector.clear()
        self.widget.tracks_layer_selector.addItem("None", None)
        for layer in self.viewer.layers:
            if isinstance(layer, (Tracks)):
                self._available_tracks.append(layer)
                self.widget.tracks_layer_selector.addItem(layer.name)
                if layer in selected_tracks:
                    self.widget.tracks_layer_selector.setCurrentText(layer.name)

    def _get_selected_labels_layers(self):
        selected_labels = []
        for i in range(self.widget.data_layer_selector.count()):
            item = self.widget.data_layer_selector.item(i)
            if item.isSelected():
                selected_labels.append(self._available_labels[i])
        return selected_labels

    def _get_selected_tracks_layers(self):
        selected_tracks = []
        for i in range(self.widget.tracks_layer_selector.count()):
            if self.widget.tracks_layer_selector.itemText(i) != "None":
                selected_tracks.append(self._available_tracks[i])
        return selected_tracks

    def _convert_selected_layer_properties_to_dataframe(self):
        selected_layer = self._get_selected_labels_layers()
        if len(selected_layer) == 0:
            self.std_out("No labels layer selected, select at least one labels layer")
            return
        if not any(isinstance(layer, Labels) for layer in selected_layer):
            self.std_out("No labels layer selected, select at least one labels layer")
            return
        if len(selected_layer) == 1 and selected_layer[0].properties:
            return pd.DataFrame(selected_layer[0].properties)

        merged_df = None
        for idx, label in enumerate(selected_layer):
            properties = label.properties
            if properties:  # Check if properties dictionary is not empty
                df = pd.DataFrame(properties)
                if not all(lf in df.columns for lf in ["label", "frame"]):
                    self.std_out(
                        f"'label' and/ or 'frame' property missing for labels layer {label.name}"
                    )
                    return

                # Rename coordinate columns with appended label name
                label_name = label.name
                df.rename(
                    columns=lambda col: f"{col}_{label_name}"
                    if col not in ["label", "frame"]
                    else col,
                    inplace=True,
                )

                # Merge the data frames on label ID
                if merged_df is None:
                    merged_df = df
                else:
                    merged_df = pd.merge(merged_df, df, on=["label", "frame"])

        if merged_df is None:
            self.std_out("No properties found")
            return

        return merged_df

    def _convert_selected_tracks_layer_data_to_dataframe(self):
        selected_layer = self._get_selected_tracks_layers()
        if selected_layer[0].data.shape[1] == 4:
            df = pd.DataFrame(
                selected_layer[0].data,
                columns=[
                    self.data_storage_instance.columns.value.object_id,
                    self.data_storage_instance.columns.value.frame_column,
                    self.data_storage_instance.columns.value.y_column,
                    self.data_storage_instance.columns.value.x_column,
                ],
            )
        elif selected_layer[0].data.shape[1] == 5:
            df = pd.DataFrame(
                selected_layer[0].data,
                columns=[
                    self.data_storage_instance.columns.value.object_id,
                    self.data_storage_instance.columns.value.frame_column,
                    str(self.data_storage_instance.columns.value.z_column),
                    self.data_storage_instance.columns.value.y_column,
                    self.data_storage_instance.columns.value.x_column,
                ],
            )
            if self.data_storage_instance.columns.value.z_column is None:
                df = df.drop(
                    columns=[str(self.data_storage_instance.columns.value.z_column)]
                )
        else:
            raise ValueError("Track layer has wrong shape")

        return df

    def _set_choices_names_from_previous(self, picker: columnpicker, col_names):
        """Sets the column names from the previous loaded data."""
        settable_columns = picker.settable_columns
        for ui_element, column_name in zip(settable_columns, col_names):
            all_items = [ui_element.itemText(i) for i in range(ui_element.count())]
            if column_name in all_items:
                ui_element.setCurrentText(column_name)

    def _run_data_loading(self, filename, delimiter=None):
        # self.loading_thread = QThread(self.widget)
        self.loading_worker = DataLoader(
            filename,
            delimiter,
            wait_for_columnpicker=True,
        )
        self.picker.rejected.connect(self._abort_loading_worker)
        self.picker.accepted.connect(self._set_loading_worker_columnpicker)

        self.widget.start_loading_icon()
        self.loading_worker.new_data.connect(self._succesfully_loaded)
        self.loading_worker.aborted.connect(self._loading_aborted)
        self.loading_worker.finished.connect(self.widget.stop_loading_icon)
        self.widget.load_data_button.setEnabled(False)
        self.widget.load_data_button.setText("")
        self.loading_worker.start()

    def closeEvent(self):
        if self.picker.isVisible():
            self.picker.close()
        try:
            self.loading_thread.quit()
            self.loading_thread.deleteLater()
            self.loading_thread.wait(1000)
        except (AttributeError, RuntimeError):
            pass
        try:
            self.matching_thread.quit()
            self.matching_thread.deleteLater()
            self.matching_thread.wait(1000)
        except (AttributeError, RuntimeError):
            pass

    def _set_loading_worker_columnpicker(self):
        self.loading_worker.wait_for_columnpicker = False

    def _abort_loading_worker(self):
        self.loading_worker.wait_for_columnpicker = False
        self.loading_worker.abort_loading = True

    def _succesfully_loaded(self, dataframe: pd.DataFrame):
        """Updates the data storage with the loaded data."""
        self._set_datastorage_to_default()
        self.data_storage_instance.columns.value = self.picker.as_columnames_object

        out_meas_name, dataframe_processed = preprocess_data(
            dataframe, self.data_storage_instance.columns.value
        )

        # set new measurmement name and data
        self.data_storage_instance.columns.value.measurement_column = out_meas_name
        self.data_storage_instance.original_data.value = dataframe_processed

    def _run_dataframe_matching(self, df1, df2, frame_col, coord_cols1):
        """Start the DataFrameMatcher in a separate thread."""
        self.matching_worker = DataFrameMatcher(
            df1, df2, frame_col=frame_col, coord_cols1=coord_cols1
        )

        # Connect signals and slots
        self.matching_worker.matched.connect(self._on_matching_success)
        self.matching_worker.aborted.connect(self._matching_aborted)

        self.widget.start_loading_icon()
        self.matching_worker.finished.connect(self.widget.stop_loading_icon)
        self.matching_worker.start()

    def _on_matching_success(self, matched_df):
        """Handle the successful completion of the DataFrameMatcher operation."""
        # Continue the processing from where you left off in _succesfully_loaded_from_layer
        out_meas_name, dataframe_processed = preprocess_data(
            matched_df, self.data_storage_instance.columns.value
        )

        # set new measurement name and data
        self.data_storage_instance.columns.value.measurement_column = out_meas_name
        self.data_storage_instance.original_data.value = dataframe_processed

    def _matching_aborted(self, error):
        """Handle the DataFrameMatcher operation abort."""
        self.std_out(f"Loading aborted by error: {error}")
        print(error)

    def _succesfully_loaded_from_layer(self, dataframe: pd.DataFrame):
        """Updates the data storage with the loaded data."""
        self._set_datastorage_to_default()
        self.data_storage_instance.columns.value = self.picker.as_columnames_object
        if (
            self.data_storage_instance.columns.value.object_id
            == "From napari tracks layer"
        ):
            self.data_storage_instance.columns.value.object_id = "track_id"

            tracks_df = self._convert_selected_tracks_layer_data_to_dataframe()
            if self.data_storage_instance.columns.value.z_column:
                coords = [
                    self.data_storage_instance.columns.value.z_column,
                    self.data_storage_instance.columns.value.y_column,
                    self.data_storage_instance.columns.value.x_column,
                ]
            else:
                coords = [
                    self.data_storage_instance.columns.value.y_column,
                    self.data_storage_instance.columns.value.x_column,
                ]

            # Start the DataFrame matching in a separate thread
            self._run_dataframe_matching(
                dataframe,
                tracks_df,
                self.data_storage_instance.columns.value.frame_column,
                coords,
            )

        else:
            self._on_matching_success(dataframe)

    def _loading_aborted(self, err_code):
        """If the loading of the data is aborted, the data storage is not updated."""
        self.widget.load_data_button.setEnabled(True)
        self.widget.load_data_button.setText("Load Data")
        if err_code == 0:
            return
        if err_code == 1:
            self.std_out("Loading aborted")
            return
        if err_code == 2:
            self.std_out("Loading aborted by error")
            return
        self.std_out(f"Loading aborted by error: {err_code}")
        print(err_code)

    def _set_datastorage_to_default(self):
        self.data_storage_instance.reset_relevant_attributes(trigger_callback=False)



"""Export image sequence from napari viewer."""

from copy import deepcopy

import numpy as np
from skimage import io
from tqdm import tqdm


def resize_napari(final_shape, viewer):
    """Iterate over window until screenshot size matches given shape.
    Center the camera and set zoom to 1 (1 canvas pixel == 1 data pixel).

    Parameters
    ----------
    final_shape : tuple
        Desired shape of the screenshot.
    viewer : napari.Viewer
        Napari viewer instance.
    """
    shape = final_shape  # init with good guess
    viewer.window.resize(shape[0], shape[1])
    for i in range(80):  # nb iterations
        current_shape = viewer.screenshot().shape[:2]
        error = np.subtract(final_shape, list(current_shape))
        error = error * 0.5  # how quick should the error be reduced
        if sum(error) == 0:
            print(f"Reached final window size after {i} iterations.")
            break
        shape = [shape[0] + error[0], shape[1] + error[1]]
        viewer.window.resize(shape[1].astype(int), shape[0].astype(int) * 2)
    viewer.reset_view()
    if len(viewer.dims.range) == 3:
        _, rgy, rgx = deepcopy(viewer.dims.range)
    if len(viewer.dims.range) == 4:
        _, _, rgy, rgx = deepcopy(viewer.dims.range)
    # Napari uses float64 for dims
    maxx, maxy = rgx[1], rgy[1]
    zoom_factor = [final_shape[0] / maxx, final_shape[1] / maxy]
    viewer.camera.zoom = min(zoom_factor)
    viewer.screenshot()


class MovieExporter:
    """Export image sequence from napari viewer."""

    def __init__(self, viewer, automatic_viewer_size, outdir, width, height):
        self.viewer = viewer
        self.automatic_viewer_size: bool = automatic_viewer_size
        self.outdir = outdir
        self.height = height
        self.width = width

    def run(self):
        """Run the exporter."""
        self._export_image_sequence(self.viewer, self.outdir)

    def _export_image_sequence(self, viewer, outdir):
        """Export a movie from a napari viewer.

        Parameters
        ----------
        viewer : napari.Viewer
            napari viewer object.
        outdir : str
            Path to output directory.
        fps : int
            Frames per second.
        """
        # Resize window to match screenshot size
        if self.automatic_viewer_size:
            resize_napari((self.width, self.height), viewer)
        # Iterate over frames
        self._iterate_over_frames(viewer, outdir)

    def _iterate_over_frames(self, viewer, outdir):
        """Iterate over frames and save screenshots.

        Parameters
        ----------
        viewer : napari.Viewer
            napari viewer object.
        outdir : str
            Path to output directory.
        """
        for frame in tqdm(range(int(viewer.dims.range[0][1]))):
            if len(viewer.dims.current_step) == 3:
                viewer.dims.current_step = (frame, 0, 0)
            elif len(viewer.dims.current_step) == 4:
                viewer.dims.current_step = (frame, 0, 0, 0)
            screenshot = viewer.screenshot(path=None, canvas_only=True, flash=False)
            io.imsave(f"{outdir}_{frame:03d}.png", screenshot, check_contrast=False)



"""Utility functions to handle data pre-processing."""

__author__ = """Benjamin Graedel"""
__email__ = "benjamin.graedel@unibe.ch"
__version__ = "0.1.3"


from arcos_gui.processing._arcos_wrapper import BatchProcessor, arcos_worker
from arcos_gui.processing._data_storage import ArcosParameters, DataStorage, columnnames
from arcos_gui.processing._preprocessing_utils import (
    DataFrameMatcher,
    DataLoader,
    create_file_names,
    create_output_folders,
    filter_data,
    get_tracklengths,
    match_dataframes,
    preprocess_data,
    process_input,
    read_data_header,
)

__all__ = [
    "DataStorage",
    "columnnames",
    "timestamp_parameters",
    "filter_data",
    "get_tracklengths",
    "process_input",
    "DataLoader",
    "read_data_header",
    "arcos_worker",
    "preprocess_data",
    "match_dataframes",
    "DataFrameMatcher",
    "create_output_folders",
    "create_file_names",
    "BatchProcessor",
    "ArcosParameters",
]



"""Utility functoins to handle data processing."""

__author__ = """Benjamin Graedel"""
__email__ = "benjamin.graedel@unibe.ch"
__version__ = "0.1.3"


from arcos_gui.tools._config import (
    ALLOWED_SETTINGS,
    ARCOS_LAYERS,
    ARCOSPARAMETERS_DEFAULTS,
    AVAILABLE_OPTIONS_FOR_BATCH,
    COLOR_CYCLE,
    OPERATOR_DICTIONARY,
    TAB20,
)
from arcos_gui.tools._image_sequence_export import MovieExporter
from arcos_gui.tools._plots import CollevPlotter, NoodlePlot, TimeSeriesPlots
from arcos_gui.tools._shape_functions import (
    fix_3d_convex_hull,
    get_bbox,
    get_bbox_3d,
    get_verticesHull,
    make_surface_3d,
    reshape_by_input_string,
)

from ._ui_util_func import (
    BatchFileDialog,
    OutputOrderValidator,
    ParameterFileDialog,
    ThrottledCallback,
    get_layer_list,
    remove_layers_after_columnpicker,
    set_track_lenths,
)

__all__ = [
    "ALLOWED_SETTINGS",
    "ARCOSPARAMETERS_DEFAULTS",
    "AVAILABLE_OPTIONS_FOR_BATCH",
    "OPERATOR_DICTIONARY",
    "COLOR_CYCLE",
    "TAB20",
    "ARCOS_LAYERS",
    "MovieExporter",
    "NoodlePlot",
    "CollevPlotter",
    "TimeSeriesPlots",
    "remove_layers_after_columnpicker",
    "get_layer_list",
    "set_track_lenths",
    "get_bbox",
    "get_bbox_3d",
    "fix_3d_convex_hull",
    "get_verticesHull",
    "make_surface_3d",
    "reshape_by_input_string",
    "ThrottledCallback",
    "BatchFileDialog",
    "OutputOrderValidator",
    "ParameterFileDialog",
]
